# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Open Buildings datasets."""

import glob
import json
import os
import sys
from collections.abc import Callable, Iterable
from typing import Any, ClassVar, cast

import fiona
import fiona.transform
import matplotlib.pyplot as plt
import pandas as pd
import rasterio
import shapely
import shapely.wkt as wkt
import torch
from matplotlib.figure import Figure
from rasterio.crs import CRS
from rtree.index import Index, Property

from .errors import DatasetNotFoundError
from .geo import VectorDataset
from .utils import BoundingBox, Path, check_integrity


class OpenBuildings(VectorDataset):
    r"""Open Buildings dataset.

    The `Open Buildings
    <https://sites.research.google/open-buildings/>`__ dataset
    consists of computer generated building detections across the African continent.

    Dataset features:

    * 516M building detections as polygons with centroid lat/long
    * covering area of 19.4M km\ :sup:`2`\  (64% of the African continent)
    * confidence score and
      `Plus Code <https://maps.google.com/pluscodes/>`_

    Dataset format:

    * csv files containing building detections compressed as csv.gz
    * meta data geojson file

    The data can be downloaded from `here
    <https://sites.research.google/open-buildings/#open-buildings-download>`__.
    Additionally, the `meta data geometry file
    <https://openbuildings-public-dot-gweb-research.uw.r.appspot.com/public/tiles.geojson>`_
    also needs to be placed in `root` as `tiles.geojson`.

    If you use this dataset in your research, please cite the following technical
    report:

    * https://arxiv.org/abs/2107.12283

    .. versionadded:: 0.3
    """

    md5s: ClassVar[dict[str, str]] = {
        '025_buildings.csv.gz': '41db2572bfd08628d01475a2ee1a2f17',
        '04f_buildings.csv.gz': '3232c1c6d45c1543260b77e5689fc8b1',
        '05b_buildings.csv.gz': '4fc57c63bbbf9a21a3902da7adc3a670',
        '093_buildings.csv.gz': '00fce146dadf0b30255e750c4c5ac2de',
        '095_buildings.csv.gz': 'f5765b0936f7ccbd0b4abed60d994f08',
        '0c3_buildings.csv.gz': '013b130fe872387e0cff842399b423de',
        '0c3_buildings.csv': 'a697ad2433e9a9f6001de25b4664651a',
        '0c5_buildings.csv.gz': '16ca283e9344e9da8b47acaf03c1c6e4',
        '0c7_buildings.csv.gz': 'b3774930006497a80c8a2fbf33056610',
        '0d1_buildings.csv.gz': '41e652218ca5964d297d9cd1d84b831c',
        '0d7_buildings.csv.gz': 'd365fe47d10b0756dd54ceca24598d8e',
        '0d9_buildings.csv.gz': '3ebd47fa4f86857266e9a7346d6aa163',
        '0db_buildings.csv.gz': '368213e9caa7ee229ef9403b0ca8c80d',
        '0dd_buildings.csv.gz': '8f5fcefff262fdfd82800092d2e9d841',
        '0df_buildings.csv.gz': 'cbb5f63b10daa25568bdde8d9f66f8a4',
        '0e1_buildings.csv.gz': 'a9b9bf1e541b62c8a34d2f6f2ae71e1c',
        '0e3_buildings.csv.gz': '3d9c2ffc11c02aec2bd008699f9c4bd1',
        '0e5_buildings.csv.gz': '1e1b2bf63dfc520e62e4b68db23fe64c',
        '0e7_buildings.csv.gz': 'c96797588c90e66268367cb56b4b9af8',
        '0e9_buildings.csv.gz': 'c53bb7bbc8140034d1be2c49ff49af68',
        '0eb_buildings.csv.gz': '407c771f614a15d69d78f1e25decf694',
        '0ed_buildings.csv.gz': 'bddd10992d291677019d7106ce1f4fac',
        '0ef_buildings.csv.gz': 'd1b91936e7ac06c661878ef9eb5dba7b',
        '0f1_buildings.csv.gz': '9d86eb10d2d8766e1385b6c52c11d5e2',
        '0f9_buildings.csv.gz': '1c6775131214b26f4a27b4c42d6e9fca',
        '0fb_buildings.csv.gz': 'd39528cb4e0cbff589ca89dc86d9b5db',
        '0fd_buildings.csv.gz': '304fe4a60e950c900697d975098f7536',
        '0ff_buildings.csv.gz': '266ca7ed1ad0251b3999b0e2e9b54648',
        '103_buildings.csv.gz': '8d3cafab5f1e02b2a0a6180eb34d1cac',
        '105_buildings.csv.gz': 'dd61cc74239aa9a1b30f10859122807b',
        '107_buildings.csv.gz': '823c05984f859a1bf17af8ce78bf2892',
        '109_buildings.csv.gz': 'cfdee0e807168cd1c183d9c01535369b',
        '10b_buildings.csv.gz': 'd8ecaf406abd864b641ba34985f3042e',
        '10d_buildings.csv.gz': 'af584a542a17942ff7e94653322dba87',
        '10f_buildings.csv.gz': '3d5369e15c4d1f59fb38cf61f4e6290b',
        '111_buildings.csv.gz': '47504e43d1b67101bed5d924225328dc',
        '113_buildings.csv.gz': '3f991c831569f91f34eaa8fc3882b2fd',
        '117_buildings.csv.gz': 'a4145fa6e458480e30c807f80ae5cd65',
        '119_buildings.csv.gz': '5661b7ac23f266542c7e0d962a8cae58',
        '11b_buildings.csv.gz': '41b6d036610d0bddac069ec72e68710e',
        '11d_buildings.csv.gz': '1ef75e9d176dd8d6bfa6012d36b1d25c',
        '11f_buildings.csv.gz': 'f004873d1ef3933c1716ab6409565b7d',
        '121_buildings.csv.gz': '0c7e7a9043ed069fbdefdcfcfc437482',
        '123_buildings.csv.gz': 'c46bd53b67025c3de11657220cce0aec',
        '125_buildings.csv.gz': '33253ae1a82656f4eedca9bd86f981a3',
        '127_buildings.csv.gz': '2f827f8fc93485572178e9ad0c65e22d',
        '129_buildings.csv.gz': '74f98346990a1d1e41241ce8f4bb201a',
        '12f_buildings.csv.gz': 'b1b0777296df2bfef512df0945ca3e14',
        '131_buildings.csv.gz': '8362825b10c9396ecbb85c49cd210bc6',
        '137_buildings.csv.gz': '96da7389df820405b0010db4a6c98c61',
        '139_buildings.csv.gz': 'c41e26fc6f3565c3d7c66ab977dc8159',
        '13b_buildings.csv.gz': '981d4ccb0f41a103bdad8ef949eb4ffe',
        '13d_buildings.csv.gz': 'd15585d06ee74b0095842dd887197035',
        '141_buildings.csv.gz': 'ae0bf17778d45119c74e50e06a04020d',
        '143_buildings.csv.gz': '9699809e57eb097dfaf9d484f1d9c5fa',
        '145_buildings.csv.gz': '81e74e0165ea358278ce18507dddfdb0',
        '147_buildings.csv.gz': '39edad15fa16c432f5d460f0a8166032',
        '149_buildings.csv.gz': '94bf8f8fa221744fb1d57c7d4065e69e',
        '14f_buildings.csv.gz': 'ca8410be89b5cf868c2a67861712e4ea',
        '15b_buildings.csv.gz': '8c0071c0ae20a60e8dd4d7aa6aac5a99',
        '15d_buildings.csv.gz': '35f044a323556adda5f31e8fc9307c85',
        '161_buildings.csv.gz': 'ba08b70a26f07b5e2cd4eafd9d6f826b',
        '163_buildings.csv.gz': '2bec83a2504b531cd1cb0311fcb6c952',
        '165_buildings.csv.gz': '48f934733dd3054164f9b09abee63312',
        '167_buildings.csv.gz': 'bba8657024d80d44e475759b65adc969',
        '169_buildings.csv.gz': '13e142e48597ee7a8b0b812e226dfa72',
        '16b_buildings.csv.gz': '9c62351d6cc8eaf761ab89d4586d26d6',
        '16d_buildings.csv.gz': 'a33c23da3f603c8c3eacc5e6a47aaf66',
        '16f_buildings.csv.gz': '4850dd7c8f0fb628ba5864ea9f47647b',
        '171_buildings.csv.gz': '4217f1b025db869c8bed1014704c2a79',
        '173_buildings.csv.gz': '5a5f3f07e261a9dc58c6180b69130e4a',
        '175_buildings.csv.gz': '5bbf7a7c8f57d28e024ddf8f4039b575',
        '177_buildings.csv.gz': '76cd4b17d68d62e1f088f229b65f8acf',
        '179_buildings.csv.gz': 'a5a1c6609483336ddff91b2385e70eb9',
        '17b_buildings.csv.gz': 'a47c1145a3b0bcdaba18c153b7b92b87',
        '17d_buildings.csv.gz': '3226d0abf396f44c1a436be83538dfd8',
        '17f_buildings.csv.gz': '3e18d4fc5837ee89274d30f2126b92b2',
        '181_buildings.csv.gz': 'c87639d7f6d6a85a3fa6b06910b0e145',
        '183_buildings.csv.gz': 'e94438ebf19b3b25035954d23a0e90cf',
        '185_buildings.csv.gz': '8de8d1d50c16c575f85b96dee474cb56',
        '189_buildings.csv.gz': 'da94cd495a99496fd687bbb4a1715c90',
        '18b_buildings.csv.gz': '9ab353335fe6ff694e834889be2b305d',
        '18d_buildings.csv.gz': 'e37e0f868ce96f7d14f7bf1a301da1d3',
        '18f_buildings.csv.gz': 'e9000b9ef9bb0f838088e96becfc95a1',
        '191_buildings.csv.gz': 'c00bb4d6b2b12615d576c06fe545cbfa',
        '193_buildings.csv.gz': 'd48d4c03ef053f6987b3e6e9e78a8b03',
        '195_buildings.csv.gz': 'd93ab833e74480f07a5ccf227067db5a',
        '197_buildings.csv.gz': '8667e040f9863e43924aafe6071fabc7',
        '199_buildings.csv.gz': '04ba65a4caf16cc1e0d5c4e1322c5885',
        '19b_buildings.csv.gz': 'e49412e3e1bccceb0bdb4df5201288f4',
        '19d_buildings.csv.gz': '92b5fb4e96529d90e99c788e3e8696d4',
        '19f_buildings.csv.gz': 'c023f6c37d0026b56f530b841517a6cd',
        '1a1_buildings.csv.gz': '471483b50c722af104af8a582e780c04',
        '1a3_buildings.csv.gz': '0a453053f1ff53f9e165e16c7f97354a',
        '1a5_buildings.csv.gz': '1f6a823e223d5f29c66aa728933de684',
        '1a7_buildings.csv.gz': '6130b724501fa16e6d84e484c4091f1f',
        '1a9_buildings.csv.gz': '73022e8e7b994e76a58cc763a057d542',
        '1b9_buildings.csv.gz': '48dea4af9d12b755e75b76c68c47de6b',
        '1bb_buildings.csv.gz': 'dfb9ee4d3843d81722b70f7582c775a4',
        '1bd_buildings.csv.gz': 'fdea2898fc50ae25b6196048373d8244',
        '1bf_buildings.csv.gz': '96ef27d6128d0bcdfa896fed6f27cdd0',
        '1c1_buildings.csv.gz': '32e3667d939e7f95316eb75a6ffdb603',
        '1c3_buildings.csv.gz': 'ed94b543da1bbe3101ed66f7d7727d24',
        '1c5_buildings.csv.gz': 'ce527ab33e564f0cc1b63ae467932a18',
        '1c7_buildings.csv.gz': 'd5fb474466d6a11d3b08e3a011984ada',
        '1dd_buildings.csv.gz': '9e7e50e3f95b3f2ceff6351b75ca1e75',
        '1e5_buildings.csv.gz': 'f95ea85fce47ce7edf5729086d43f922',
        '1e7_buildings.csv.gz': '2bca5682c48134e69b738d70dfe7d516',
        '1e9_buildings.csv.gz': 'f049ad06dbbb200f524b4f50d1df8c2e',
        '1eb_buildings.csv.gz': '6822d7f202b453ec3cc03fb8f04691ad',
        '1ed_buildings.csv.gz': '9dfc560e2c3d135ebdcd46fa09c47169',
        '1ef_buildings.csv.gz': '506e7772c35b09cfd3b6f8691dc2947d',
        '1f1_buildings.csv.gz': 'b74f2b585cfad3b881fe4f124080440a',
        '1f3_buildings.csv.gz': '12896642315320e11ed9ed2d3f0e5995',
        '1f5_buildings.csv.gz': '334aea21e532e178bf5c54d028158906',
        '1f7_buildings.csv.gz': '0e8c3d2e005eb04c6852a8aa993f5a76',
        '217_buildings.csv.gz': '296e9ba121fea752b865a48e5c0fe8a5',
        '219_buildings.csv.gz': '1d19b6626d738f7706f75c2935aaaff4',
        '21d_buildings.csv.gz': '28bfca1f8668f59db021d3a195994768',
        '21f_buildings.csv.gz': '06325c8b0a8f6ed598b7dc6f0bb5adf2',
        '221_buildings.csv.gz': 'a354ffc1f7226d525c7cf53848975da1',
        '223_buildings.csv.gz': '3bda1339d561b3bc749220877f1384d9',
        '225_buildings.csv.gz': '8eb02ad77919d9e551138a14d3ad1bbc',
        '227_buildings.csv.gz': 'c07aceb7c81f83a653810befa0695b61',
        '22f_buildings.csv.gz': '97d63e30e008ec4424f6b0641b75377c',
        '231_buildings.csv.gz': 'f4bc384ed74552ddcfe2e69107b91345',
        '233_buildings.csv.gz': '081756e7bdcfdc2aee9114c4cfe62bd8',
        '23b_buildings.csv.gz': '75776d3dcbc90cf3a596664747880134',
        '23d_buildings.csv.gz': 'e5d0b9b7b14601f58cfdb9ea170e9520',
        '23f_buildings.csv.gz': '77f38466419b4d391be8e4f05207fdf5',
        '3d1_buildings.csv.gz': '6659c97bd765250b0dee4b1b7ff583a9',
        '3d5_buildings.csv.gz': 'c27d8f6b2808549606f00bc04d8b42bc',
        '3d7_buildings.csv.gz': 'abdef2e68cc31c67dbb6e60c4c40483e',
        '3d9_buildings.csv.gz': '4c06ae37d8e76626345a52a32f989de9',
        '3db_buildings.csv.gz': 'e83ca0115eaf4ec0a72aaf932b00442a',
        'b5b_buildings.csv.gz': '5e5f59cb17b81137d89c4bab8107e837',
    }

    filename_glob = '*_buildings.csv'
    zipfile_glob = '*_buildings.csv.gz'

    meta_data_url = 'https://sites.research.google/open-buildings/tiles.geojson'
    meta_data_filename = 'tiles.geojson'

    def __init__(
        self,
        paths: Path | Iterable[Path] = 'data',
        crs: CRS | None = None,
        res: float | tuple[float, float] = 0.0001,
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Dataset instance.

        Args:
            paths: one or more root directories to search or files to load
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS in (xres, yres) format. If a
                single float is provided, it is used for both the x and y resolution.
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found.

        .. versionchanged:: 0.5
           *root* was renamed to *paths*.
        """
        self.paths = paths
        if isinstance(res, int | float):
            res = (res, res)
        self.res = res
        self.checksum = checksum
        self.transforms = transforms

        self._verify()

        # Create an R-tree to index the dataset using the polygon centroid as bounds
        self.index = Index(interleaved=False, properties=Property(dimension=3))

        assert isinstance(self.paths, str | os.PathLike)
        with open(os.path.join(self.paths, 'tiles.geojson')) as f:
            data = json.load(f)

        features = data['features']
        features_filenames = [
            feature['properties']['tile_url'].split('/')[-1] for feature in features
        ]  # get csv filename

        polygon_files = glob.glob(os.path.join(self.paths, self.zipfile_glob))
        polygon_filenames = [f.split(os.sep)[-1] for f in polygon_files]

        matched_features = [
            feature
            for filename, feature in zip(features_filenames, features)
            if filename in polygon_filenames
        ]

        i = 0
        source_crs = CRS.from_dict({'init': 'epsg:4326'})
        for feature in matched_features:
            if crs is None:
                crs = CRS.from_dict(source_crs)

            c = feature['geometry']['coordinates'][0]
            xs = [x[0] for x in c]
            ys = [x[1] for x in c]

            minx, miny, maxx, maxy = min(xs), min(ys), max(xs), max(ys)

            (minx, maxx), (miny, maxy) = fiona.transform.transform(
                source_crs.to_dict(), crs.to_dict(), [minx, maxx], [miny, maxy]
            )
            mint = 0
            maxt = sys.maxsize
            coords = (minx, maxx, miny, maxy, mint, maxt)

            filepath = os.path.join(
                self.paths, feature['properties']['tile_url'].split('/')[-1]
            )
            self.index.insert(i, coords, filepath)
            i += 1

        if i == 0:
            raise DatasetNotFoundError(self)

        self._crs = crs
        self._source_crs = source_crs

    def __getitem__(self, query: BoundingBox) -> dict[str, Any]:
        """Retrieve image/mask and metadata indexed by query.

        Args:
            query: (minx, maxx, miny, maxy, mint, maxt) coordinates to index

        Returns:
            sample of image/mask and metadata for the given query. If there are
            not matching shapes found within the query, an empty raster is returned

        Raises:
            IndexError: if query is not found in the index
        """
        hits = self.index.intersection(tuple(query), objects=True)
        filepaths = cast(list[str], [hit.object for hit in hits])

        if not filepaths:
            raise IndexError(
                f'query: {query} not found in index with bounds: {self.bounds}'
            )

        shapes = self._filter_geometries(query, filepaths)

        # Rasterize geometries
        width = (query.maxx - query.minx) / self.res[0]
        height = (query.maxy - query.miny) / self.res[1]
        transform = rasterio.transform.from_bounds(
            query.minx, query.miny, query.maxx, query.maxy, width, height
        )
        if shapes:
            masks = rasterio.features.rasterize(
                shapes, out_shape=(round(height), round(width)), transform=transform
            )
            masks = torch.tensor(masks).unsqueeze(0)
        else:
            masks = torch.zeros(size=(1, round(height), round(width)))

        sample = {'mask': masks, 'crs': self.crs, 'bounds': query}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _filter_geometries(
        self, query: BoundingBox, filepaths: list[str]
    ) -> list[dict[str, Any]]:
        """Filters a df read from the polygon csv file based on query and conf thresh.

        Args:
            query: (minx, maxx, miny, maxy, mint, maxt) coordinates to index
            filepaths: filepaths to files that were hits from rmtree index

        Returns:
            List with all polygons from all hit filepaths

        """
        # We need to know the bounding box of the query in the source CRS
        (minx, maxx), (miny, maxy) = fiona.transform.transform(
            self._crs.to_dict(),
            self._source_crs.to_dict(),
            [query.minx, query.maxx],
            [query.miny, query.maxy],
        )
        df_query = (
            f'longitude >= {minx} & longitude <= {maxx} & '
            f'latitude >= {miny} & latitude <= {maxy}'
        )
        shapes = []
        for f in filepaths:
            csv_chunks = pd.read_csv(f, chunksize=200000, compression='gzip')
            for chunk in csv_chunks:
                df = chunk.query(df_query)
                # Warp geometries to requested CRS
                polygon_series = df['geometry'].map(self._wkt_fiona_geom_transform)
                shapes.extend(polygon_series.values.tolist())

        return shapes

    def _wkt_fiona_geom_transform(self, x: str) -> dict[str, Any]:
        """Function to transform a geometry string into new crs.

        Args:
            x: Polygon string

        Returns:
            transformed geometry in geojson format

        """
        x = json.dumps(shapely.geometry.mapping(wkt.loads(x)))
        x = json.loads(x.replace("'", '"'))
        import fiona

        if hasattr(fiona, 'model'):
            import fiona.model

            geom = fiona.model.Geometry(**x)
        else:
            geom = x
        transformed: dict[str, Any] = fiona.transform.transform_geom(
            self._source_crs.to_dict(), self._crs.to_dict(), geom
        )
        return transformed

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the zip files have already been downloaded and checksum
        assert isinstance(self.paths, str | os.PathLike)
        pathname = os.path.join(self.paths, self.zipfile_glob)
        i = 0
        for zipfile in glob.iglob(pathname):
            filename = os.path.basename(zipfile)
            if self.checksum and not check_integrity(zipfile, self.md5s[filename]):
                raise RuntimeError(f'Dataset found, but corrupted: {filename}.')
            i += 1

        if i != 0:
            return

        raise DatasetNotFoundError(self)

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        mask = sample['mask'].permute(1, 2, 0)

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            pred = sample['prediction'].permute(1, 2, 0)
            ncols = 2
        else:
            ncols = 1

        fig, axs = plt.subplots(nrows=1, ncols=ncols, figsize=(ncols * 4, 4))

        if showing_predictions:
            axs[0].imshow(mask)
            axs[0].axis('off')
            axs[1].imshow(pred)
            axs[1].axis('off')
            if show_titles:
                axs[0].set_title('Mask')
                axs[1].set_title('Prediction')
        else:
            axs.imshow(mask)
            axs.axis('off')
            if show_titles:
                axs.set_title('Mask')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""ZueriCrop dataset."""

import os
from collections.abc import Callable, Sequence

import matplotlib.pyplot as plt
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError, RGBBandsMissingError
from .geo import NonGeoDataset
from .utils import Path, download_url, lazy_import, percentile_normalization


class ZueriCrop(NonGeoDataset):
    """ZueriCrop dataset.

    The `ZueriCrop <https://github.com/0zgur0/multi-stage-convSTAR-network>`__
    dataset is a dataset for time-series instance segmentation of crops.

    Dataset features:

    * Sentinel-2 multispectral imagery
    * instance masks of 48 crop categories
    * nine multispectral bands
    * 116k images with 10 m per pixel resolution (24x24 px)
    * ~28k time-series containing 142 images each

    Dataset format:

    * single hdf5 dataset containing images, semantic masks, and instance masks
    * data is parsed into images and instance masks, boxes, and labels
    * one mask per time-series

    Dataset classes:

    * 48 fine-grained hierarchical crop `categories
      <https://github.com/0zgur0/multi-stage-convSTAR-network/blob/fa92b5b3cb77f5171c5c3be740cd6e6395cc29b6/labels.csv>`_

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1016/j.rse.2021.112603

    .. note::

       This dataset requires the following additional library to be installed:

       * `h5py <https://pypi.org/project/h5py/>`_ to load the dataset
    """

    url = 'https://hf.co/datasets/torchgeo/zuericrop/resolve/8ac0f416fbaab032d8670cc55f984b9f079e86b2/'
    md5s = ('1635231df67f3d25f4f1e62c98e221a4', '5118398c7a5bbc246f5f6bb35d8d529b')
    filenames = ('ZueriCrop.hdf5', 'labels.csv')

    band_names = ('NIR', 'B03', 'B02', 'B04', 'B05', 'B06', 'B07', 'B11', 'B12')
    rgb_bands = ('B04', 'B03', 'B02')

    def __init__(
        self,
        root: Path = 'data',
        bands: Sequence[str] = band_names,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new ZueriCrop dataset instance.

        Args:
            root: root directory where dataset can be found
            bands: the subset of bands to load
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
            DependencyNotFoundError: If h5py is not installed.
        """
        lazy_import('h5py')

        self._validate_bands(bands)
        self.band_indices = torch.tensor(
            [self.band_names.index(b) for b in bands]
        ).long()

        self.root = root
        self.bands = bands
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self.filepath = os.path.join(root, 'ZueriCrop.hdf5')

        self._verify()

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            sample containing image, mask, bounding boxes, and target label
        """
        image = self._load_image(index)
        mask, boxes, label = self._load_target(index)

        sample = {'image': image, 'mask': mask, 'bbox_xyxy': boxes, 'label': label}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        h5py = lazy_import('h5py')
        with h5py.File(self.filepath, 'r') as f:
            length: int = f['data'].shape[0]
        return length

    def _load_image(self, index: int) -> Tensor:
        """Load a single image.

        Args:
            index: index to return

        Returns:
            the image
        """
        h5py = lazy_import('h5py')
        with h5py.File(self.filepath, 'r') as f:
            array = f['data'][index, ...]

        tensor = torch.from_numpy(array)
        # Convert from TxHxWxC to TxCxHxW
        tensor = tensor.permute((0, 3, 1, 2))
        tensor = torch.index_select(tensor, dim=1, index=self.band_indices)
        return tensor

    def _load_target(self, index: int) -> tuple[Tensor, Tensor, Tensor]:
        """Load the target mask for a single image.

        Args:
            index: index to return

        Returns:
            the target mask and label for each mask
        """
        import h5py

        with h5py.File(self.filepath, 'r') as f:
            mask_array = f['gt'][index, ...]
            instance_array = f['gt_instance'][index, ...]

        mask_tensor = torch.from_numpy(mask_array)
        instance_tensor = torch.from_numpy(instance_array)

        # Convert from HxWxC to CxHxW
        mask_tensor = mask_tensor.permute((2, 0, 1))
        instance_tensor = instance_tensor.permute((2, 0, 1))

        # Convert instance mask of N instances to N binary instance masks
        instance_ids = torch.unique(instance_tensor)
        # Exclude a mask for unknown/background
        instance_ids = instance_ids[instance_ids != 0]
        instance_ids = instance_ids[:, None, None]
        masks = instance_tensor == instance_ids

        # Parse labels for each instance
        labels_list = []
        for mask in masks:
            label = mask_tensor[mask[None, :, :]]
            label = torch.unique(label)[0]
            labels_list.append(label)

        # Get bounding boxes for each instance
        boxes_list = []
        for mask in masks:
            pos = torch.where(mask)
            xmin = torch.min(pos[1])
            xmax = torch.max(pos[1])
            ymin = torch.min(pos[0])
            ymax = torch.max(pos[0])
            boxes_list.append([xmin, ymin, xmax, ymax])

        masks = masks.to(torch.uint8)
        boxes = torch.tensor(boxes_list).to(torch.float)
        labels = torch.tensor(labels_list).to(torch.long)

        return masks, boxes, labels

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        exists = []
        for filename in self.filenames:
            filepath = os.path.join(self.root, filename)
            exists.append(os.path.exists(filepath))

        if all(exists):
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        for filename, md5 in zip(self.filenames, self.md5s):
            filepath = os.path.join(self.root, filename)
            if not os.path.exists(filepath):
                download_url(
                    self.url + filename,
                    self.root,
                    filename=filename,
                    md5=md5 if self.checksum else None,
                )

    def _validate_bands(self, bands: Sequence[str]) -> None:
        """Validate list of bands.

        Args:
            bands: user-provided sequence of bands to load

        Raises:
            AssertionError: if ``bands`` is not a sequence
            ValueError: if an invalid band name is provided

        .. versionadded:: 0.2
        """
        assert isinstance(bands, Sequence), "'bands' must be a sequence"
        for band in bands:
            if band not in self.band_names:
                raise ValueError(f"'{band}' is an invalid band name.")

    def plot(
        self,
        sample: dict[str, Tensor],
        time_step: int = 0,
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            time_step: time step at which to access image, beginning with 0
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            RGBBandsMissingError: If *bands* does not include all RGB bands.

        .. versionadded:: 0.2
        """
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise RGBBandsMissingError()

        ncols = 2
        image, mask = sample['image'][time_step, rgb_indices], sample['mask']

        image = torch.tensor(
            percentile_normalization(image.numpy()) * 255, dtype=torch.uint8
        )

        mask = torch.argmax(mask, dim=0)

        if 'prediction' in sample:
            ncols += 1
            preds = torch.argmax(sample['prediction'], dim=0)

        fig, axs = plt.subplots(ncols=ncols, figsize=(10, 10 * ncols))

        axs[0].imshow(image.permute(1, 2, 0))
        axs[0].axis('off')
        axs[1].imshow(mask)
        axs[1].axis('off')

        if show_titles:
            axs[0].set_title('Image')
            axs[1].set_title('Mask')

        if 'prediction' in sample:
            axs[2].imshow(preds)
            axs[2].axis('off')
            if show_titles:
                axs[2].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# https://github.com/zhu-xlab/Copernicus-FM

"""Copernicus Foundation Model (Copernicus-FM)."""

import math
from collections.abc import Sequence
from typing import Any, Literal

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from einops import rearrange
from timm.models.vision_transformer import Block
from torch import Tensor, vmap
from torchvision.models._api import Weights, WeightsEnum

from ..samplers.utils import _to_tuple
from .dofa import FCResLayer, TransformerWeightGenerator


def resize_abs_pos_embed(
    pos_embed: Tensor,
    new_size: int | tuple[int, int],
    old_size: int | tuple[int, int],
    num_prefix_tokens: int = 1,
    interpolation: str = 'bicubic',
    antialias: bool = True,
) -> Tensor:
    """Resize absolute position embeddings to a target resolution via interpolation.

    Adapted from https://github.com/bwconrad/flexivit. Copyright (c) 2023 Ben Conrad.

    Args:
        pos_embed: Position embeddings tensor of size [b, n, d]
        new_size: Target [height, width] of embedding
        old_size: Original [height, width] of embedding
        num_prefix_tokens: Number of non-spatial prefix tokens (e.g., cls)
        interpolation: Resize interpolation type
        antialias: Whether to apply antialiasing resizing

    Returns:
        Resized pos_embed of size [b, n', d]
    """
    new_size = _to_tuple(new_size)
    old_size = _to_tuple(old_size)
    new_ntok = new_size[0] * new_size[1]

    # Return if no resize necessary
    if new_size == old_size:
        return pos_embed

    if num_prefix_tokens:
        posemb_prefix, pos_embed = (
            pos_embed[:, :num_prefix_tokens],
            pos_embed[:, num_prefix_tokens:],
        )
    else:
        posemb_prefix, pos_embed = None, pos_embed

    # Interpolate position embedding
    pos_embed = pos_embed.reshape(1, old_size[0], old_size[1], -1).permute(0, 3, 1, 2)
    pos_embed = F.interpolate(
        pos_embed, size=new_size, mode=interpolation, antialias=antialias
    )
    pos_embed = pos_embed.permute(0, 2, 3, 1).reshape(1, new_ntok, -1)

    # Add back extra prefix tokens
    if posemb_prefix is not None:
        pos_embed = torch.cat([posemb_prefix, pos_embed], dim=1)

    return pos_embed


def pi_resize_patch_embed(
    patch_embed: Tensor,
    new_patch_size: tuple[int, int],
    interpolation: str = 'bicubic',
    antialias: bool = True,
) -> Tensor:
    """Resample patch embeddings to a target resolution via pseudo-inverse resizing.

    Adapted from https://github.com/bwconrad/flexivit. Copyright (c) 2023 Ben Conrad.

    Args:
        patch_embed: Patch embedding parameters of size [d, c, h, w]
        new_patch_size: Target [height, width] of embedding
        interpolation: Resize interpolation type
        antialias: Whether to apply antialiasing resizing

    Returns:
        Resized pos_embed of size [d, c h', w']
    """
    assert len(patch_embed.shape) == 4, 'Patch embed kernel should be a 4D tensor'
    assert len(new_patch_size) == 2, 'New patch size should only be (height, width)'

    _, _, h, w = patch_embed.shape
    old_patch_size = (h, w)

    # Return original kernel if no resize is necessary
    if old_patch_size == new_patch_size:
        return patch_embed

    def resize(x: Tensor, shape: tuple[int, int]) -> Tensor:
        x = F.interpolate(
            x[None, None, ...], shape, mode=interpolation, antialias=antialias
        )
        return x[0, 0, ...]

    def calculate_pinv(
        old_shape: tuple[int, int], new_shape: tuple[int, int]
    ) -> Tensor:
        mat = []
        for i in range(np.prod(old_shape)):
            basis_vec = torch.zeros(old_shape)
            basis_vec[np.unravel_index(i, old_shape)] = 1.0
            mat.append(resize(basis_vec, new_shape).reshape(-1))
        resize_matrix = torch.stack(mat)
        pinv: Tensor = torch.linalg.pinv(resize_matrix)
        return pinv

    # Calculate pseudo-inverse of resize matrix
    resize_matrix_pinv = calculate_pinv(old_patch_size, new_patch_size)
    resize_matrix_pinv = resize_matrix_pinv.to(patch_embed.device)

    def resample_patch_embed(patch_embed: Tensor) -> Tensor:
        h, w = new_patch_size
        resampled_kernel = resize_matrix_pinv @ patch_embed.reshape(-1)
        return rearrange(resampled_kernel, '(h w) -> h w', h=h, w=w)

    v_resample_patch_embed = vmap(vmap(resample_patch_embed, 0, 0), 1, 1)

    patch_embed = v_resample_patch_embed(patch_embed)
    return patch_embed


class FourierExpansion(nn.Module):
    """A Fourier series-style expansion into a high-dimensional space.

    Adapted from https://github.com/microsoft/aurora.
    Copyright (c) Microsoft Corporation.
    """

    def __init__(self, lower: float, upper: float, assert_range: bool = True) -> None:
        """Initialise.

        Args:
            lower: Lower wavelength.
            upper: Upper wavelength.
            assert_range: Assert that the encoded tensor is within the specified
                wavelength range.
        """
        super().__init__()
        self.lower = lower
        self.upper = upper
        self.assert_range = assert_range

    def forward(self, x: Tensor, d: int) -> Tensor:
        """Perform the expansion.

        Adds a dimension of length `d` to the end of the shape of `x`.

        Args:
            x: Input to expand of shape `(..., n)`. All elements of `x` must
               lie within `[self.lower, self.upper]` if `self.assert_range` is `True`.
            d: Dimensionality. Must be a multiple of two.

        Raises:
            AssertionError: If `self.assert_range` is `True` and not all elements of `x`
                are not within `[self.lower, self.upper]`.
            ValueError: If `d` is not a multiple of two.

        Returns:
            Fourier series-style expansion of `x` of shape `(..., n, d)`.
        """
        # If the input is not within the configured range, the embedding might be ambiguous!
        in_range = torch.logical_and(
            self.lower <= x.abs(), torch.all(x.abs() <= self.upper)
        )
        # Allow zeros to pass through.
        in_range_or_zero = torch.all(torch.logical_or(in_range, x == 0))
        if self.assert_range and not in_range_or_zero:
            raise AssertionError(
                f'The input tensor is not within the configured range'
                f' `[{self.lower}, {self.upper}]`.'
            )

        # We will use half of the dimensionality for `sin` and the other half for `cos`.
        if not (d % 2 == 0):
            raise ValueError('The dimensionality must be a multiple of two.')

        # Always perform the expansion with `float64`s to avoid numerical accuracy shenanigans.
        x = x.double()

        wavelengths = torch.logspace(
            math.log10(self.lower),
            math.log10(self.upper),
            d // 2,
            base=10,
            device=x.device,
            dtype=x.dtype,
        )
        prod = torch.einsum('...i,j->...ij', x, 2 * np.pi / wavelengths)
        encoding = torch.cat((torch.sin(prod), torch.cos(prod)), dim=-1)

        return encoding.float()  # Cast to `float32` to avoid incompatibilities.


class DynamicPatchEmbed(nn.Module):
    """Dynamic patch embedding with spectral or variable hypernetworks.

    Adapted from DOFA.
    """

    def __init__(
        self,
        hyper_dim: int = 128,
        kernel_size: int = 16,
        embed_dim: int = 1024,
        input_mode: Literal['spectral', 'variable'] = 'spectral',
    ) -> None:
        """Initialize a new DynamicPatchEmbed instance.

        Args:
            hyper_dim: dim for wavelength/bandwidth/varname encoding.
            kernel_size: Kernel size for the patch embedding (convolution) layer.
            embed_dim: Embedding dimension.
            input_mode: Type of hypernetwork to use. Options: 'spectral' or 'variable'.
                'spectral' uses Fourier encodings for wavelength and bandwidth;
                'variable' uses a language embedding for variable names.
        """
        super().__init__()
        self.input_mode = input_mode
        self.kernel_size = kernel_size
        self.hyper_dim = hyper_dim
        self.embed_dim = embed_dim
        self._num_kernel = self.kernel_size * self.kernel_size * self.embed_dim
        self.patch_size = (kernel_size, kernel_size)
        self.num_patches = -1

        if self.input_mode == 'spectral':
            # Spectral hypernetwork: Fourier encoding for wavelength and bandwidth.
            # min wavelength: ultraviolet light (100 nm)
            # max wavelength: radio waves (1 m)
            # min bandwidth: s2 ~ 10 nm
            # max bandwidth: s1 ~ 1 m
            self.spectrum_central_expansion = FourierExpansion(100, 1e9)
            self.spectrum_bandwidth_expansion = FourierExpansion(1, 1e9)
        elif self.input_mode == 'variable':
            # Variable hypernetwork: Language embedding for variable names.
            self.language_proj = nn.Linear(2048, self.hyper_dim)

        self.weight_generator = TransformerWeightGenerator(
            hyper_dim, self._num_kernel, embed_dim
        )
        self.scaler = 0.01
        self.fclayer = FCResLayer(hyper_dim)
        self._init_weights()

    def _get_weights(self, waves: Tensor) -> Tensor:
        """Use the dynamic weight generator.

        Args:
            waves: Spectral wavelengths.

        Returns:
            Dynamic weights.
        """
        dynamic_weights: Tensor = self.weight_generator(waves)
        return dynamic_weights

    def weight_init(self, m: object) -> None:
        """Initialize weights of a single layer.

        Args:
            m: A single layer.
        """
        if isinstance(m, nn.Linear):
            init.xavier_uniform_(m.weight)
            if m.bias is not None:
                m.bias.data.fill_(0.01)

    def _init_weights(self) -> None:
        """Initialize weights of all layers."""
        self.weight_generator.apply(self.weight_init)
        self.fclayer.apply(self.weight_init)

    def forward(
        self,
        x: Tensor,
        wavelengths: Tensor | None = None,
        bandwidths: Tensor | None = None,
        language_embed: Tensor | None = None,
        kernel_size: int | None = None,
    ) -> Tensor:
        """Forward pass.

        For input_mode=='spectral', `wavelengths` and `bandwidths` must be provided.
        For input_mode=='variable', `language_embed` must be provided.

        Args:
            x: Input image tensor (B, C, H, W).
            wavelengths: Wavelengths in nm (required if input_mode=='spectral').
            bandwidths: Bandwidths in nm (required if input_mode=='spectral').
            language_embed: Language embedding tensor from Llama 3.2 1B (length 2048).
            kernel_size: If provided and differs from the initialized kernel size,
                the generated patch embed kernel weights are resized accordingly.

        Returns:
            Output after patch embedding (B, N, D).

        Raises:
            ValueError: When *input_mode=='spectral'* and *wavelengths* or *bandwidths* is missing,
                or when *input_mode=='variable'* and *language_embed* is missing.
        """
        if self.input_mode == 'spectral':
            if wavelengths is None or bandwidths is None:
                msg = 'For spectral hypernet, wavelengths and bandwidths must be provided.'
                raise ValueError(msg)

            emb_central = self.spectrum_central_expansion(wavelengths, self.hyper_dim)
            emb_bandwidth = self.spectrum_bandwidth_expansion(
                bandwidths, self.hyper_dim
            )
            waves = emb_central + emb_bandwidth
        elif self.input_mode == 'variable':
            if language_embed is None:
                msg = 'For variable hypernet, language_embed must be provided.'
                raise ValueError(msg)

            # Expand dims to match batch size.
            waves = self.language_proj(language_embed.unsqueeze(0))

        waves = self.fclayer(waves)
        weight, bias = self._get_weights(waves)
        inplanes = waves.size(0)
        dynamic_weight = weight.view(
            inplanes, self.kernel_size, self.kernel_size, self.embed_dim
        )
        dynamic_weight = dynamic_weight.permute(3, 0, 1, 2)

        if kernel_size is not None and self.kernel_size != kernel_size:
            dynamic_weight = pi_resize_patch_embed(
                dynamic_weight, (kernel_size, kernel_size)
            )
        else:
            kernel_size = self.kernel_size

        if bias is not None:
            bias = bias.view(self.embed_dim) * self.scaler

        weights = dynamic_weight * self.scaler

        dynamic_out = F.conv2d(
            x, weights, bias=bias, stride=kernel_size, padding=1, dilation=1
        )
        x = dynamic_out.flatten(2).transpose(1, 2)
        return x


class CopernicusFM(nn.Module):
    """CopernicusFM: VisionTransformer backbone.

    Example:
        **1. Spectral Mode (Using Wavelength and Bandwidth):**

        >>> model = CopernicusFM()
        >>> x = torch.randn(1, 4, 224, 224) # input image
        >>> metadata = torch.full((1, 4), float('nan')) # [lon (degree), lat (degree), delta_time (days since 1970/1/1), patch_token_area (km^2)], assume unknown
        >>> wavelengths = [490, 560, 665, 842] # wavelength (nm): B,G,R,NIR (Sentinel 2)
        >>> bandwidths = [65, 35, 30, 115] # bandwidth (nm): B,G,R,NIR (Sentinel 2)
        >>> kernel_size = 16 # expected patch size
        >>> input_mode = 'spectral'
        >>> logit = model(x, metadata, wavelengths=wavelengths, bandwidths=bandwidths, input_mode=input_mode, kernel_size=kernel_size)
        >>> print(logit.shape)

        **2. Variable Mode (Using language embedding):**

        >>> model = CopernicusFM()
        >>> varname = 'Sentinel 5P Nitrogen Dioxide' # variable name (as input to a LLM for language embed)
        >>> x = torch.randn(1, 1, 56, 56) # input image
        >>> metadata = torch.full((1, 4), float('nan')) # [lon (degree), lat (degree), delta_time (days since 1970/1/1), patch_token_area (km^2)], assume unknown
        >>> language_embed = torch.randn(2048) # language embedding: encode varname with a LLM (e.g. Llama)
        >>> kernel_size = 4 # expected patch size
        >>> input_mode = 'variable'
        >>> logit = model(x, metadata, language_embed=language_embed, input_mode=input_mode, kernel_size=kernel_size)
        >>> print(logit.shape)

    """

    def __init__(
        self,
        img_size: int = 224,
        patch_size: int = 16,
        drop_rate: float = 0.0,
        embed_dim: int = 1024,
        depth: int = 24,
        num_heads: int = 16,
        hyper_dim: int = 128,
        num_classes: int = 0,
        global_pool: bool = True,
        mlp_ratio: float = 4.0,
        norm_layer: type[nn.Module] = nn.LayerNorm,
    ) -> None:
        """Initialize a new CopernicusFM instance.

        Args:
            img_size: Input image size.
            patch_size: Patch size.
            drop_rate: Head dropout rate.
            embed_dim: Transformer embedding dimension.
            depth: Depth of transformer.
            num_heads: Number of attention heads.
            hyper_dim: Dimensions of dynamic weight generator.
            num_classes: Number of classes for classification head.
            global_pool: Whether or not to perform global pooling.
            mlp_ratio: Ratio of MLP hidden dim to embedding dim.
            norm_layer: Normalization layer.
        """
        super().__init__()

        self.hyper_dim = hyper_dim
        self.global_pool = global_pool
        if self.global_pool:
            norm_layer = norm_layer
            embed_dim = embed_dim
            self.fc_norm = norm_layer(embed_dim)
        else:
            self.norm = norm_layer(embed_dim)

        self.patch_embed_spectral = DynamicPatchEmbed(
            hyper_dim=hyper_dim,
            kernel_size=patch_size,
            embed_dim=embed_dim,
            input_mode='spectral',
        )
        self.patch_embed_variable = DynamicPatchEmbed(
            hyper_dim=hyper_dim,
            kernel_size=patch_size,
            embed_dim=embed_dim,
            input_mode='variable',
        )

        self.num_patches = (img_size // patch_size) ** 2

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))

        # Fixed sin-cos embedding
        self.pos_embed = nn.Parameter(
            torch.zeros(1, self.num_patches + 1, embed_dim), requires_grad=False
        )

        self.coord_expansion = FourierExpansion(0.0001, 720)
        self.scale_expansion = FourierExpansion(0.001, 5.1e8)  # 1 m2 to 5.1e8 km2
        # 1 to 365.25 days, enable more than 1 year
        self.time_expansion = FourierExpansion(1, 365.25, assert_range=False)
        self.coord_fc = nn.Linear(embed_dim, embed_dim)
        self.scale_fc = nn.Linear(embed_dim, embed_dim)
        self.time_fc = nn.Linear(embed_dim, embed_dim)
        # if metadata is not available, set to a learned parameter
        self.coord_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.scale_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.time_token = nn.Parameter(torch.zeros(1, 1, embed_dim))

        self.blocks = nn.ModuleList(
            [
                Block(
                    embed_dim,
                    num_heads,
                    mlp_ratio,
                    qkv_bias=True,
                    norm_layer=norm_layer,
                )
                for i in range(depth)
            ]
        )

        self.head_drop = nn.Dropout(drop_rate)
        self.head = (
            nn.Linear(embed_dim, num_classes) if num_classes > 0 else nn.Identity()
        )

    def get_coord_pos_embed(self, lons: Tensor, lats: Tensor, embed_dim: int) -> Tensor:
        """Geospatial coordinate position embedding.

        Args:
            lons: Longitudes (x).
            lats: Latitudes (y).
            embed_dim: Embedding dimension.

        Returns:
            Coordinate position embedding.
        """
        coord_embed_lon = self.coord_expansion(lons + 180, embed_dim // 2)
        coord_embed_lat = self.coord_expansion(lats + 90, embed_dim // 2)
        coord_embed = torch.cat([coord_embed_lon, coord_embed_lat], dim=-1)

        if coord_embed.shape[-1] < embed_dim:
            # pad zeros
            coord_embed = torch.cat(
                (
                    coord_embed,
                    torch.zeros(
                        coord_embed.shape[0],
                        embed_dim - coord_embed.shape[-1],
                        device=coord_embed.device,
                    ),
                ),
                dim=-1,
            )

        return coord_embed.unsqueeze(1)  # [B,1,D]

    def get_area_pos_embed(self, areas: Tensor, embed_dim: int) -> Tensor:
        """Geospatial area position embedding.

        Args:
            areas: Spatial areas.
            embed_dim: Embedding dimension.

        Returns:
            Area position embedding.
        """
        scale_embed: Tensor = self.scale_expansion(areas, embed_dim)  # B, D
        scale_embed = scale_embed.unsqueeze(1)  # [B,1,D]
        return scale_embed

    def get_time_pos_embed(self, times: Tensor, embed_dim: int) -> Tensor:
        """Geotemporal position embedding.

        Args:
            times: Timestamps.
            embed_dim: Embedding dimension.

        Returns:
            Temporal position embedding.
        """
        time_embed: Tensor = self.time_expansion(times, embed_dim)  # B, D
        time_embed = time_embed.unsqueeze(1)  # [B,1,D]
        return time_embed

    def forward_features(
        self,
        x: Tensor,
        metadata: Tensor,
        wavelengths: Sequence[float] | None = None,
        bandwidths: Sequence[float] | None = None,
        language_embed: Tensor | None = None,
        input_mode: Literal['spectral', 'variable'] = 'spectral',
        kernel_size: int | None = None,
    ) -> Tensor:
        """Forward pass of the feature embedding layer.

        Args:
            x: Input mini-batch.
            metadata: Longitudes (degree), latitudes (degree), times
                (days since 1970/1/1), and areas (km^2) of each patch.
                Use NaN for unknown metadata.
            wavelengths: Wavelengths of each spectral band (nm).
                Only used if *input_mode=='spectral'*.
            bandwidths: Bandwidths in nm.
                Only used if *input_mode=='spectral'*.
            language_embed: Language embedding tensor from Llama 3.2 1B (length 2048).
                Only used if *input_mode=='variable'*.
            input_mode: One of 'spectral' or 'variable'.
            kernel_size: If provided and differs from the initialized kernel size,
                the generated patch embed kernel weights are resized accordingly.

        Returns:
            Output mini-batch.
        """
        if input_mode == 'spectral':
            wvs = torch.tensor(wavelengths, device=x.device).float()
            bws = torch.tensor(bandwidths, device=x.device).float()
            x = self.patch_embed_spectral(
                x, wavelengths=wvs, bandwidths=bws, kernel_size=kernel_size
            )
        elif input_mode == 'variable':
            x = self.patch_embed_variable(
                x, language_embed=language_embed, kernel_size=kernel_size
            )

        # resize pos embed
        num_patches = x.size(1)
        num_patches_sqrt = int(math.sqrt(num_patches))
        num_patches_sqrt_origin = int(math.sqrt(self.num_patches))
        pos_embed = resize_abs_pos_embed(
            self.pos_embed,
            num_patches_sqrt,
            (num_patches_sqrt_origin, num_patches_sqrt_origin),
            num_prefix_tokens=1,
        )

        # coord, scale and time pos embed
        lons, lats, times, areas = (
            metadata[:, 0],
            metadata[:, 1],
            metadata[:, 2],
            metadata[:, 3],
        )
        embed_dim = pos_embed.shape[-1]
        if torch.isnan(lons).any() or torch.isnan(lats).any():
            coord_embed: nn.Parameter | Tensor = self.coord_token
        else:
            coord_embed = self.get_coord_pos_embed(lons, lats, embed_dim)
        coord_embed = self.coord_fc(coord_embed)
        if torch.isnan(areas).any():
            area_embed: nn.Parameter | Tensor = self.scale_token
        else:
            area_embed = self.get_area_pos_embed(areas, embed_dim)
        area_embed = self.scale_fc(area_embed)
        if torch.isnan(times).any():
            time_embed: nn.Parameter | Tensor = self.time_token
        else:
            time_embed = self.get_time_pos_embed(times, embed_dim)
        time_embed = self.time_fc(time_embed)
        pos_embed = pos_embed + coord_embed + area_embed + time_embed

        # add pos embed w/o cls token
        x = x + pos_embed[:, 1:, :]

        # append cls token
        cls_token = self.cls_token + pos_embed[:, :1, :]
        cls_tokens = cls_token.expand(x.shape[0], -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)

        # apply Transformer blocks
        for block in self.blocks:
            x = block(x)

        if self.global_pool:
            x = x[:, 1:, :].mean(dim=1)  # global pool without cls token
            outcome: Tensor = self.fc_norm(x)
        else:
            x = self.norm(x)
            outcome = x[:, 0]
        return outcome

    def forward_head(self, x: Tensor, pre_logits: bool = False) -> Tensor:
        """Forward pass of the attention head.

        Args:
            x: Input mini-batch.
            pre_logits: Whether or not to return the layer before logits are computed.

        Returns:
            Output mini-batch.
        """
        x = self.head_drop(x)
        return x if pre_logits else self.head(x)

    def forward(
        self,
        x: Tensor,
        metadata: Tensor,
        wavelengths: Sequence[float] | None = None,
        bandwidths: Sequence[float] | None = None,
        language_embed: Tensor | None = None,
        input_mode: Literal['spectral', 'variable'] = 'spectral',
        kernel_size: int | None = None,
    ) -> Tensor:
        """Forward pass of the model.

        Args:
            x: Input mini-batch.
            metadata: Longitudes (degree), latitudes (degree), times
                (days since 1970/1/1), and areas (km^2) of each patch.
                Use NaN for unknown metadata.
            wavelengths: Wavelengths of each spectral band (nm).
                Only used if *input_mode=='spectral'*.
            bandwidths: Bandwidths in nm.
                Only used if *input_mode=='spectral'*.
            language_embed: Language embedding tensor from Llama 3.2 1B (length 2048).
                Only used if *input_mode=='variable'*.
            input_mode: One of 'spectral' or 'variable'.
            kernel_size: If provided and differs from the initialized kernel size,
                the generated patch embed kernel weights are resized accordingly.

        Returns:
            Output mini-batch.
        """
        fx = self.forward_features(
            x,
            metadata,
            wavelengths,
            bandwidths,
            language_embed,
            input_mode,
            kernel_size,
        )
        x = self.forward_head(fx)
        return x


class CopernicusFM_Base_Weights(WeightsEnum):  # type: ignore[misc]
    """Copernicus-FM-base weights."""

    CopernicusFM_ViT = Weights(
        url='https://huggingface.co/torchgeo/copernicus-fm/resolve/f395812cc990ba25a451dbb9c9e6d95c8482947e/CopernicusFM_ViT_base_varlang-085350e4.pth',
        transforms=None,
        meta={
            'dataset': 'Copernicus-Pretrain',
            'model': 'copernicusfm_base',
            'publication': 'https://arxiv.org/abs/2503.11849',
            'repo': 'https://github.com/zhu-xlab/Copernicus-FM',
            'ssl_method': 'mae+distill',
        },
    )


def copernicusfm_base(
    weights: CopernicusFM_Base_Weights | None = None, *args: Any, **kwargs: Any
) -> CopernicusFM:
    """CopernicusFM vit-base model.

    If you use this model in your research, please cite the following paper:

    * https://arxiv.org/abs/2503.11849

    .. versionadded:: 0.7

    Args:
        weights: Pre-trained model weights to use.
        *args: Additional arguments to pass to :class:`CopernicusFM`.
        **kwargs: Additional keyword arguments to pass to :class:`CopernicusFM`.

    Returns:
        A CopernicusFM base model.
    """
    kwargs |= {'embed_dim': 768, 'depth': 12, 'num_heads': 12}
    model = CopernicusFM(*args, **kwargs)

    if weights:
        missing_keys, unexpected_keys = model.load_state_dict(
            weights.get_state_dict(progress=True), strict=False
        )

        # Both fc_norm and head are generated dynamically
        assert set(missing_keys) <= {
            'fc_norm.weight',
            'fc_norm.bias',
            'head.weight',
            'head.bias',
        }
        assert not unexpected_keys

    return model



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""OSCD datamodule."""

from typing import Any

import kornia.augmentation as K
import torch
from torch import Tensor
from torch.utils.data import random_split

from ..datasets import OSCD
from ..samplers.utils import _to_tuple
from ..transforms.transforms import _ExtractPatches
from .geo import NonGeoDataModule

MEAN = {
    'B01': 1565.696044921875,
    'B02': 1351.3319091796875,
    'B03': 1257.1082763671875,
    'B04': 1254.932861328125,
    'B05': 1388.689208984375,
    'B06': 1827.6710205078125,
    'B07': 2050.2744140625,
    'B08': 1963.4619140625,
    'B8A': 2182.680908203125,
    'B09': 629.837646484375,
    'B10': 14.855598449707031,
    'B11': 1909.8394775390625,
    'B12': 1379.6024169921875,
}

STD = {
    'B01': 263.7977600097656,
    'B02': 394.5567321777344,
    'B03': 508.9673767089844,
    'B04': 726.4053344726562,
    'B05': 686.6111450195312,
    'B06': 730.0204467773438,
    'B07': 822.0133056640625,
    'B08': 842.5917358398438,
    'B8A': 895.7645263671875,
    'B09': 314.8407287597656,
    'B10': 9.417905807495117,
    'B11': 984.9249267578125,
    'B12': 844.7711181640625,
}


class OSCDDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the OSCD dataset.

    Uses the train/test splits from the dataset and further splits
    the train split into train/val splits.

    .. versionadded:: 0.2
    """

    def __init__(
        self,
        batch_size: int = 32,
        patch_size: tuple[int, int] | int = 64,
        val_split_pct: float = 0.2,
        num_workers: int = 0,
        **kwargs: Any,
    ) -> None:
        """Initialize a new OSCDDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            patch_size: Size of each patch, either ``size`` or ``(height, width)``.
                Should be a multiple of 32 for most segmentation architectures.
            val_split_pct: Percentage of the dataset to use as a validation set.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.OSCD`.
        """
        super().__init__(OSCD, batch_size=batch_size, num_workers=num_workers, **kwargs)

        self.patch_size = _to_tuple(patch_size)
        self.val_split_pct = val_split_pct

        self.bands = kwargs.get('bands', OSCD.all_bands)
        self.mean = torch.tensor([MEAN[b] for b in self.bands])
        self.std = torch.tensor([STD[b] for b in self.bands])

        self.train_aug = K.AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std),
            K.RandomCrop(self.patch_size, pad_if_needed=True),
            data_keys=None,
            keepdim=True,
        )
        self.aug = K.AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std),
            _ExtractPatches(window_size=self.patch_size),
            data_keys=None,
            keepdim=True,
            same_on_batch=True,
        )

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        if stage in ['fit', 'validate']:
            self.dataset = OSCD(split='train', **self.kwargs)
            generator = torch.Generator().manual_seed(0)
            self.train_dataset, self.val_dataset = random_split(
                self.dataset, [1 - self.val_split_pct, self.val_split_pct], generator
            )
        if stage in ['test']:
            self.test_dataset = OSCD(split='test', **self.kwargs)

    def on_after_batch_transfer(
        self, batch: dict[str, Tensor], dataloader_idx: int
    ) -> dict[str, Tensor]:
        """Apply batch augmentations to the batch after it is transferred to the device.

        Args:
            batch: A batch of data that needs to be altered or augmented.
            dataloader_idx: The index of the dataloader to which the batch belongs.

        Returns:
            A batch of data.

        .. versionadded:: 0.7
        """
        # This solves a special case where if batch_size=1 the mask won't be stacked correctly
        if batch['mask'].ndim == 3:
            batch['mask'] = batch['mask'].unsqueeze(dim=0)

        batch = super().on_after_batch_transfer(batch, dataloader_idx)
        batch['mask'] = batch['mask'].squeeze(dim=1)
        return batch



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Trainers for image classification."""

import os
from typing import Any, Literal

import kornia.augmentation as K
import matplotlib.pyplot as plt
import timm
import torch.nn as nn
from matplotlib.figure import Figure
from segmentation_models_pytorch.losses import FocalLoss, JaccardLoss
from torch import Tensor
from torchmetrics import MetricCollection
from torchmetrics.classification import Accuracy, FBetaScore, JaccardIndex
from torchvision.models._api import WeightsEnum
from typing_extensions import deprecated

from ..datasets import RGBBandsMissingError, unbind_samples
from ..models import get_weight
from . import utils
from .base import BaseTask


class ClassificationTask(BaseTask):
    """Image classification."""

    def __init__(
        self,
        model: str = 'resnet50',
        weights: WeightsEnum | str | bool | None = None,
        in_channels: int = 3,
        task: Literal['binary', 'multiclass', 'multilabel'] = 'multiclass',
        num_classes: int | None = None,
        num_labels: int | None = None,
        loss: Literal['ce', 'bce', 'jaccard', 'focal'] = 'ce',
        class_weights: Tensor | None = None,
        lr: float = 1e-3,
        patience: int = 10,
        freeze_backbone: bool = False,
    ) -> None:
        """Initialize a new ClassificationTask instance.

        Args:
            model: Name of the `timm
                <https://huggingface.co/docs/timm/reference/models>`__ model to use.
            weights: Initial model weights. Either a weight enum, the string
                representation of a weight enum, True for ImageNet weights, False
                or None for random weights, or the path to a saved model state dict.
            in_channels: Number of input channels to model.
            task: One of 'binary', 'multiclass', or 'multilabel'.
            num_classes: Number of prediction classes (only for ``task='multiclass'``).
            num_labels: Number of prediction labels (only for ``task='multilabel'``).
            loss: One of 'ce', 'bce', 'jaccard', or 'focal'.
            class_weights: Optional rescaling weight given to each
                class and used with 'ce' loss.
            lr: Learning rate for optimizer.
            patience: Patience for learning rate scheduler.
            freeze_backbone: Freeze the backbone network to linear probe
                the classifier head.

        .. versionadded:: 0.7
           The *task* and *num_labels* parameters.

        .. versionadded:: 0.5
           The *class_weights* and *freeze_backbone* parameters.

        .. versionchanged:: 0.5
           *learning_rate* and *learning_rate_schedule_patience* were renamed to
           *lr* and *patience*.

        .. versionchanged:: 0.4
           *classification_model* was renamed to *model*.
        """
        self.weights = weights
        super().__init__()

    def configure_models(self) -> None:
        """Initialize the model."""
        weights = self.weights

        # Create model
        self.model = timm.create_model(
            self.hparams['model'],
            num_classes=self.hparams['num_classes'] or self.hparams['num_labels'] or 1,
            in_chans=self.hparams['in_channels'],
            pretrained=weights is True,
        )

        # Load weights
        if weights and weights is not True:
            if isinstance(weights, WeightsEnum):
                state_dict = weights.get_state_dict(progress=True)
            elif os.path.exists(weights):
                _, state_dict = utils.extract_backbone(weights)
            else:
                state_dict = get_weight(weights).get_state_dict(progress=True)
            utils.load_state_dict(self.model, state_dict)

        # Freeze backbone and unfreeze classifier head
        if self.hparams['freeze_backbone']:
            for param in self.model.parameters():
                param.requires_grad = False
            for param in self.model.get_classifier().parameters():
                param.requires_grad = True

    def configure_losses(self) -> None:
        """Initialize the loss criterion."""
        match self.hparams['loss']:
            case 'ce':
                self.criterion: nn.Module = nn.CrossEntropyLoss(
                    weight=self.hparams['class_weights']
                )
            case 'bce':
                self.criterion = nn.BCEWithLogitsLoss()
            case 'jaccard':
                self.criterion = JaccardLoss(mode=self.hparams['task'])
            case 'focal':
                self.criterion = FocalLoss(mode=self.hparams['task'], normalized=True)

    def configure_metrics(self) -> None:
        """Initialize the performance metrics.

        * :class:`~torchmetrics.Accuracy`: The number of
          true positives divided by the dataset size. Both overall accuracy (OA)
          using 'micro' averaging and average accuracy (AA) using 'macro' averaging
          are reported. Higher values are better.
        * :class:`~torchmetrics.JaccardIndex`: Intersection
          over union (IoU). Uses 'macro' averaging. Higher valuers are better.
        * :class:`~torchmetrics.FBetaScore`: F1 score.
          The harmonic mean of precision and recall. Uses 'micro' averaging.
          Higher values are better.

        .. note::
           * 'Micro' averaging suits overall performance evaluation but may not reflect
             minority class accuracy.
           * 'Macro' averaging gives equal weight to each class, and is useful for
             balanced performance assessment across imbalanced classes.
        """
        kwargs = {
            'task': self.hparams['task'],
            'num_classes': self.hparams['num_classes'],
            'num_labels': self.hparams['num_labels'],
        }
        metrics = MetricCollection(
            {
                'OverallAccuracy': Accuracy(average='micro', **kwargs),
                'AverageAccuracy': Accuracy(average='macro', **kwargs),
                'JaccardIndex': JaccardIndex(**kwargs),
                'F1Score': FBetaScore(beta=1.0, average='micro', **kwargs),
            }
        )
        self.train_metrics = metrics.clone(prefix='train_')
        self.val_metrics = metrics.clone(prefix='val_')
        self.test_metrics = metrics.clone(prefix='test_')

    def training_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> Tensor:
        """Compute the training loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            The loss tensor.
        """
        x = batch['image']
        y = batch['label']
        batch_size = x.shape[0]
        y_hat = self(x).squeeze(1)
        self.train_metrics(y_hat, y)
        self.log_dict(self.train_metrics, batch_size=batch_size)

        if self.hparams['loss'] == 'bce':
            y = y.float()

        loss: Tensor = self.criterion(y_hat, y)
        self.log('train_loss', loss, batch_size=batch_size)

        return loss

    def validation_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        """Compute the validation loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch['image']
        y = batch['label']
        batch_size = x.shape[0]
        y_hat = self(x).squeeze(1)
        self.val_metrics(y_hat, y)
        self.log_dict(self.val_metrics, batch_size=batch_size)

        if self.hparams['loss'] == 'bce':
            y = y.float()

        loss = self.criterion(y_hat, y)
        self.log('val_loss', loss, batch_size=batch_size)

        if (
            batch_idx < 10
            and hasattr(self.trainer, 'datamodule')
            and hasattr(self.trainer.datamodule, 'plot')
            and self.logger
            and hasattr(self.logger, 'experiment')
            and hasattr(self.logger.experiment, 'add_figure')
        ):
            datamodule = self.trainer.datamodule
            aug = K.AugmentationSequential(
                K.Denormalize(datamodule.mean, datamodule.std),
                data_keys=None,
                keepdim=True,
            )
            batch = aug(batch)
            match self.hparams['task']:
                case 'binary' | 'multilabel':
                    batch['prediction'] = (y_hat.sigmoid() >= 0.5).long()
                case 'multiclass':
                    batch['prediction'] = y_hat.argmax(dim=1)

            for key in ['image', 'label', 'prediction']:
                batch[key] = batch[key].cpu()
            sample = unbind_samples(batch)[0]

            fig: Figure | None = None
            try:
                fig = datamodule.plot(sample)
            except RGBBandsMissingError:
                pass

            if fig:
                summary_writer = self.logger.experiment
                summary_writer.add_figure(
                    f'image/{batch_idx}', fig, global_step=self.global_step
                )
                plt.close()

    def test_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> None:
        """Compute the test loss and additional metrics.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.
        """
        x = batch['image']
        y = batch['label']
        batch_size = x.shape[0]
        y_hat = self(x).squeeze(1)
        self.test_metrics(y_hat, y)
        self.log_dict(self.test_metrics, batch_size=batch_size)

        if self.hparams['loss'] == 'bce':
            y = y.float()

        loss = self.criterion(y_hat, y)
        self.log('test_loss', loss, batch_size=batch_size)

    def predict_step(
        self, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> Tensor:
        """Compute the predicted class probabilities.

        Args:
            batch: The output of your DataLoader.
            batch_idx: Integer displaying index of this batch.
            dataloader_idx: Index of the current dataloader.

        Returns:
            Output predicted probabilities.
        """
        x = batch['image']
        y_hat: Tensor = self(x)

        match self.hparams['task']:
            case 'binary' | 'multilabel':
                y_hat = y_hat.sigmoid()
            case 'multiclass':
                y_hat = y_hat.softmax(dim=1)

        return y_hat


@deprecated('Use torchgeo.trainers.ClassificationTask instead')
class MultiLabelClassificationTask(ClassificationTask):
    """Multi-label image classification."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Wrapper around torchgeo.trainers.ClassificationTask to massage kwargs."""
        kwargs['task'] = 'multilabel'
        kwargs['num_labels'] = kwargs['num_classes']
        super().__init__(*args, **kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""USAVars dataset."""

import glob
import os
from collections.abc import Callable, Sequence
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_url, extract_archive


class USAVars(NonGeoDataset):
    """USAVars dataset.

    The USAVars dataset is reproduction of the dataset used in the paper "`A
    generalizable and accessible approach to machine learning with global satellite
    imagery <https://doi.org/10.1038/s41467-021-24638-z>`_". Specifically, this dataset
    includes 1 sq km. crops of NAIP imagery resampled to 4m/px cenetered on ~100k points
    that are sampled randomly from the contiguous states in the USA. Each point contains
    three continuous valued labels (taken from the dataset released in the paper): tree
    cover percentage, elevation, and population density.

    Dataset format:

    * images are 4-channel GeoTIFFs
    * labels are singular float values

    Dataset labels:

    * tree cover
    * elevation
    * population density

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1038/s41467-021-24638-z

    .. versionadded:: 0.3
    """

    data_url = 'https://hf.co/datasets/torchgeo/usavars/resolve/01377abfaf50c0cc8548aaafb79533666bbf288f/{}'
    dirname = 'uar'

    md5 = '677e89fd20e5dd0fe4d29b61827c2456'

    label_urls: ClassVar[dict[str, str]] = {
        'housing': data_url.format('housing.csv'),
        'income': data_url.format('income.csv'),
        'roads': data_url.format('roads.csv'),
        'nightlights': data_url.format('nightlights.csv'),
        'population': data_url.format('population.csv'),
        'elevation': data_url.format('elevation.csv'),
        'treecover': data_url.format('treecover.csv'),
    }

    split_metadata: ClassVar[dict[str, dict[str, str]]] = {
        'train': {
            'url': data_url.format('train_split.txt'),
            'filename': 'train_split.txt',
            'md5': '3f58fffbf5fe177611112550297200e7',
        },
        'val': {
            'url': data_url.format('val_split.txt'),
            'filename': 'val_split.txt',
            'md5': 'bca7183b132b919dec0fc24fb11662a0',
        },
        'test': {
            'url': data_url.format('test_split.txt'),
            'filename': 'test_split.txt',
            'md5': '97bb36bc003ae0bf556a8d6e8f77141a',
        },
    }

    ALL_LABELS = ('treecover', 'elevation', 'population')

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        labels: Sequence[str] = ALL_LABELS,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new USAVars dataset instance.

        Args:
            root: root directory where dataset can be found
            split: train/val/test split to load
            labels: list of labels to include
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if invalid labels are provided
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        self.root = root

        assert split in self.split_metadata
        self.split = split

        for lab in labels:
            assert lab in self.ALL_LABELS

        self.labels = labels
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        self.files = self._load_files()

        self.label_dfs = {
            lab: pd.read_csv(os.path.join(self.root, lab + '.csv'), index_col='ID')
            for lab in self.labels
        }

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        tif_file = self.files[index]
        id_ = tif_file[5:-4]

        sample = {
            'labels': Tensor(
                [self.label_dfs[lab].loc[id_][lab] for lab in self.labels]
            ),
            'image': self._load_image(os.path.join(self.root, 'uar', tif_file)),
            'centroid_lat': Tensor([self.label_dfs[self.labels[0]].loc[id_]['lat']]),
            'centroid_lon': Tensor([self.label_dfs[self.labels[0]].loc[id_]['lon']]),
        }

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.files)

    def _load_files(self) -> list[str]:
        """Loads file names."""
        with open(os.path.join(self.root, f'{self.split}_split.txt')) as f:
            files = f.read().splitlines()
        return files

    def _load_image(self, path: Path) -> Tensor:
        """Load a single image.

        Args:
            path: path to the image

        Returns:
            the image
        """
        with rasterio.open(path) as f:
            array: np.typing.NDArray[np.int_] = f.read()
            tensor = torch.from_numpy(array).float()
            return tensor

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        pathname = os.path.join(self.root, 'uar')
        csv_pathname = os.path.join(self.root, '*.csv')
        split_pathname = os.path.join(self.root, '*_split.txt')

        csv_split_count = (len(glob.glob(csv_pathname)), len(glob.glob(split_pathname)))
        if glob.glob(pathname) and csv_split_count == (7, 3):
            return

        # Check if the zip files have already been downloaded
        pathname = os.path.join(self.root, self.dirname + '.zip')
        if glob.glob(pathname) and csv_split_count == (7, 3):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset."""
        for f_name in self.label_urls:
            download_url(self.label_urls[f_name], self.root, filename=f_name + '.csv')

        download_url(self.data_url, self.root, md5=self.md5 if self.checksum else None)

        for metadata in self.split_metadata.values():
            download_url(
                metadata['url'],
                self.root,
                md5=metadata['md5'] if self.checksum else None,
            )

    def _extract(self) -> None:
        """Extract the dataset."""
        extract_archive(os.path.join(self.root, self.dirname + '.zip'))

    def plot(
        self,
        sample: dict[str, Tensor],
        show_labels: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_labels: flag indicating whether to show labels above panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        image = sample['image'][:3].numpy()  # get RGB inds
        image = np.moveaxis(image, 0, 2)

        fig, axs = plt.subplots(figsize=(10, 10))
        axs.imshow(image)
        axs.axis('off')

        if show_labels:
            labels = [(lab, val) for lab, val in sample.items() if lab != 'image']
            label_string = ''
            for lab, val in labels:
                label_string += f'{lab}={round(val[0].item(), 2)} '
            axs.set_title(label_string)

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""RESISC45 dataset."""

import os
from collections.abc import Callable
from typing import ClassVar, cast

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoClassificationDataset
from .utils import Path, download_url, extract_archive


class RESISC45(NonGeoClassificationDataset):
    """NWPU-RESISC45 dataset.

    The `RESISC45 <https://doi.org/10.1109/jproc.2017.2675998>`__
    dataset is a dataset for remote sensing image scene classification.

    Dataset features:

    * 31,500 images with 0.2-30 m per pixel resolution (256x256 px)
    * three spectral bands - RGB
    * 45 scene classes, 700 images per class
    * images extracted from Google Earth from over 100 countries
    * images conditions with high variability (resolution, weather, illumination)

    Dataset format:

    * images are three-channel jpgs

    Dataset classes:

    0. airplane
    1. airport
    2. baseball_diamond
    3. basketball_court
    4. beach
    5. bridge
    6. chaparral
    7. church
    8. circular_farmland
    9. cloud
    10. commercial_area
    11. dense_residential
    12. desert
    13. forest
    14. freeway
    15. golf_course
    16. ground_track_field
    17. harbor
    18. industrial_area
    19. intersection
    20. island
    21. lake
    22. meadow
    23. medium_residential
    24. mobile_home_park
    25. mountain
    26. overpass
    27. palace
    28. parking_lot
    29. railway
    30. railway_station
    31. rectangular_farmland
    32. river
    33. roundabout
    34. runway
    35. sea_ice
    36. ship
    37. snowberg
    38. sparse_residential
    39. stadium
    40. storage_tank
    41. tennis_court
    42. terrace
    43. thermal_power_station
    44. wetland

    This dataset uses the train/val/test splits defined in the "In-domain representation
    learning for remote sensing" paper:

    * https://arxiv.org/abs/1911.06721

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1109/jproc.2017.2675998
    """

    url = 'https://hf.co/datasets/torchgeo/resisc45/resolve/a826b44d938a883185f11ebe3d512d38b464312f/NWPU-RESISC45.zip'
    md5 = '75206b2e16446591afa88e2628744886'
    filename = 'NWPU-RESISC45.zip'
    directory = 'NWPU-RESISC45'

    splits = ('train', 'val', 'test')
    split_urls: ClassVar[dict[str, str]] = {
        'train': 'https://hf.co/datasets/torchgeo/resisc45/resolve/a826b44d938a883185f11ebe3d512d38b464312f/resisc45-train.txt',
        'val': 'https://hf.co/datasets/torchgeo/resisc45/resolve/a826b44d938a883185f11ebe3d512d38b464312f/resisc45-val.txt',
        'test': 'https://hf.co/datasets/torchgeo/resisc45/resolve/a826b44d938a883185f11ebe3d512d38b464312f/resisc45-test.txt',
    }
    split_md5s: ClassVar[dict[str, str]] = {
        'train': 'b5a4c05a37de15e4ca886696a85c403e',
        'val': 'a0770cee4c5ca20b8c32bbd61e114805',
        'test': '3dda9e4988b47eb1de9f07993653eb08',
    }

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new RESISC45 dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train", "val", or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.splits
        self.root = root
        self.download = download
        self.checksum = checksum
        self._verify()

        valid_fns = set()
        with open(os.path.join(self.root, f'resisc45-{split}.txt')) as f:
            for fn in f:
                valid_fns.add(fn.strip())

        def is_in_split(x: Path) -> bool:
            return os.path.basename(x) in valid_fns

        super().__init__(
            root=os.path.join(root, self.directory),
            transforms=transforms,
            is_valid_file=is_in_split,
        )

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        filepath = os.path.join(self.root, self.directory)
        if os.path.exists(filepath):
            return

        # Check if zip file already exists (if so then extract)
        filepath = os.path.join(self.root, self.filename)
        if os.path.exists(filepath):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download and extract the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset."""
        download_url(
            self.url,
            self.root,
            filename=self.filename,
            md5=self.md5 if self.checksum else None,
        )
        for split in self.splits:
            download_url(
                self.split_urls[split],
                self.root,
                filename=f'resisc45-{split}.txt',
                md5=self.split_md5s[split] if self.checksum else None,
            )

    def _extract(self) -> None:
        """Extract the dataset."""
        filepath = os.path.join(self.root, self.filename)
        extract_archive(filepath)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`NonGeoClassificationDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        .. versionadded:: 0.2
        """
        image = np.rollaxis(sample['image'].numpy(), 0, 3)
        label = cast(int, sample['label'].item())
        label_class = self.classes[label]

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = cast(int, sample['prediction'].item())
            prediction_class = self.classes[prediction]

        fig, ax = plt.subplots(figsize=(4, 4))
        ax.imshow(image)
        ax.axis('off')
        if show_titles:
            title = f'Label: {label_class}'
            if showing_predictions:
                title += f'\nPrediction: {prediction_class}'
            ax.set_title(title)

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""CMS Global Mangrove Canopy dataset."""

import os
from collections.abc import Callable
from typing import Any

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from rasterio.crs import CRS

from .errors import DatasetNotFoundError
from .geo import RasterDataset
from .utils import Path, check_integrity, extract_archive


class CMSGlobalMangroveCanopy(RasterDataset):
    """CMS Global Mangrove Canopy dataset.

    The `CMS Global Mangrove Canopy dataset
    <https://daac.ornl.gov/cgi-bin/dsviewer.pl?ds_id=1665>`_
    consists of a single band map at 30m resolution of either aboveground biomass (agb),
    basal area weighted height (hba95), or maximum canopy height (hmax95).

    The dataset needs to be manually downloaded from the above link, where you can make
    an account and subsequently download the dataset.

    .. versionadded:: 0.3
    """

    is_image = False

    filename_regex = r"""^
        (?P<mangrove>[A-Za-z]{8})
        _(?P<variable>[a-z0-9]*)
        _(?P<country>[A-Za-z][^.]*)
    """

    zipfile = 'CMS_Global_Map_Mangrove_Canopy_1665.zip'
    md5 = '3e7f9f23bf971c25e828b36e6c5496e3'

    all_countries = (
        'AndamanAndNicobar',
        'Angola',
        'Anguilla',
        'AntiguaAndBarbuda',
        'Aruba',
        'Australia',
        'Bahamas',
        'Bahrain',
        'Bangladesh',
        'Barbados',
        'Belize',
        'Benin',
        'Brazil',
        'BritishVirginIslands',
        'Brunei',
        'Cambodia',
        'Cameroon',
        'CarribeanCaymanIslands',
        'China',
        'Colombia',
        'Comoros',
        'CostaRica',
        'Cote',
        'CoteDivoire',
        'CotedIvoire',
        'Cuba',
        'DemocraticRepublicOfCongo',
        'Djibouti',
        'DominicanRepublic',
        'EcuadorWithGalapagos',
        'Egypt',
        'ElSalvador',
        'EquatorialGuinea',
        'Eritrea',
        'EuropaIsland',
        'Fiji',
        'Fiji2',
        'FrenchGuiana',
        'FrenchGuyana',
        'FrenchPolynesia',
        'Gabon',
        'Gambia',
        'Ghana',
        'Grenada',
        'Guadeloupe',
        'Guam',
        'Guatemala',
        'Guinea',
        'GuineaBissau',
        'Guyana',
        'Haiti',
        'Hawaii',
        'Honduras',
        'HongKong',
        'India',
        'Indonesia',
        'Iran',
        'Jamaica',
        'Japan',
        'Kenya',
        'Liberia',
        'Macau',
        'Madagascar',
        'Malaysia',
        'Martinique',
        'Mauritania',
        'Mayotte',
        'Mexico',
        'Micronesia',
        'Mozambique',
        'Myanmar',
        'NewCaledonia',
        'NewZealand',
        'Newzealand',
        'Nicaragua',
        'Nigeria',
        'NorthernMarianaIslands',
        'Oman',
        'Pakistan',
        'Palau',
        'Panama',
        'PapuaNewGuinea',
        'Peru',
        'Philipines',
        'PuertoRico',
        'Qatar',
        'ReunionAndMauritius',
        'SaintKittsAndNevis',
        'SaintLucia',
        'SaintVincentAndTheGrenadines',
        'Samoa',
        'SaudiArabia',
        'Senegal',
        'Seychelles',
        'SierraLeone',
        'Singapore',
        'SolomonIslands',
        'Somalia',
        'Somalia2',
        'Soudan',
        'SouthAfrica',
        'SriLanka',
        'Sudan',
        'Suriname',
        'Taiwan',
        'Tanzania',
        'Thailand',
        'TimorLeste',
        'Togo',
        'Tonga',
        'TrinidadAndTobago',
        'TurksAndCaicosIslands',
        'Tuvalu',
        'UnitedArabEmirates',
        'UnitedStates',
        'Vanuatu',
        'Venezuela',
        'Vietnam',
        'VirginIslandsUs',
        'WallisAndFutuna',
        'Yemen',
    )

    measurements = ('agb', 'hba95', 'hmax95')

    def __init__(
        self,
        paths: Path | list[Path] = 'data',
        crs: CRS | None = None,
        res: float | tuple[float, float] | None = None,
        measurement: str = 'agb',
        country: str = all_countries[0],
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        cache: bool = True,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Dataset instance.

        Args:
            paths: one or more root directories to search or files to load
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS in (xres, yres) format. If a
                single float is provided, it is used for both the x and y resolution.
                (defaults to the resolution of the first file found)
            measurement: which of the three measurements, 'agb', 'hba95', or 'hmax95'
            country: country for which to retrieve data
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            cache: if True, cache file handle to speed up repeated sampling
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if country or measurement arg are not str or invalid
            DatasetNotFoundError: If dataset is not found.

        .. versionchanged:: 0.5
           *root* was renamed to *paths*.
        """
        self.paths = paths
        self.checksum = checksum

        assert isinstance(country, str), 'Country argument must be a str.'
        assert country in self.all_countries, (
            f'You have selected an invalid country, please choose one of {self.all_countries}'
        )
        self.country = country

        assert isinstance(measurement, str), 'Measurement must be a string.'
        assert measurement in self.measurements, (
            f'You have entered an invalid measurement, please choose one of {self.measurements}.'
        )
        self.measurement = measurement

        self.filename_glob = f'**/Mangrove_{self.measurement}_{self.country}*'

        self._verify()

        super().__init__(paths, crs, res, transforms=transforms, cache=cache)

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        if self.files:
            return

        # Check if the zip file has already been downloaded
        assert isinstance(self.paths, str | os.PathLike)
        pathname = os.path.join(self.paths, self.zipfile)
        if os.path.exists(pathname):
            if self.checksum and not check_integrity(pathname, self.md5):
                raise RuntimeError('Dataset found, but corrupted.')
            self._extract()
            return

        raise DatasetNotFoundError(self)

    def _extract(self) -> None:
        """Extract the dataset."""
        assert isinstance(self.paths, str | os.PathLike)
        pathname = os.path.join(self.paths, self.zipfile)
        extract_archive(pathname)

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`RasterDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        mask = sample['mask'].squeeze()
        ncols = 1

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            pred = sample['prediction'].squeeze()
            ncols = 2

        fig, axs = plt.subplots(nrows=1, ncols=ncols, figsize=(ncols * 4, 4))

        if showing_predictions:
            axs[0].imshow(mask)
            axs[0].axis('off')
            axs[1].imshow(pred)
            axs[1].axis('off')
            if show_titles:
                axs[0].set_title('Mask')
                axs[1].set_title('Prediction')
        else:
            axs.imshow(mask)
            axs.axis('off')
            if show_titles:
                axs.set_title('Mask')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""MapInWild dataset."""

import os
import shutil
from collections import defaultdict
from collections.abc import Callable
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import (
    Path,
    check_integrity,
    download_url,
    extract_archive,
    percentile_normalization,
)


class MapInWild(NonGeoDataset):
    """MapInWild dataset.

    The `MapInWild <https://ieeexplore.ieee.org/document/10089830>`__ dataset is
    curated for the task of wilderness mapping on a pixel-level. MapInWild is a
    multi-modal dataset and comprises various geodata acquired and formed from
    different RS sensors over 1018 locations: dual-pol Sentinel-1, four-season
    Sentinel-2 with 10 bands, ESA WorldCover map, and Visible Infrared Imaging
    Radiometer Suite NightTime Day/Night band. The dataset consists of 8144
    images with the shape of 1920 x 1920 pixels. The images are weakly annotated
    from the World Database of Protected Areas (WDPA).

    Dataset features:

    * 1018 areas globally sampled from the WDPA
    * 10-Band Sentinel-2
    * Dual-pol Sentinel-1
    * ESA WorldCover Land Cover
    * Visible Infrared Imaging Radiometer Suite NightTime Day/Night Band

    If you use this dataset in your research, please cite the following paper:

    * https://ieeexplore.ieee.org/document/10089830

    .. versionadded:: 0.5
    """

    url = 'https://hf.co/datasets/burakekim/mapinwild/resolve/d963778e31e7e0ed2329c0f4cbe493be532f0e71/'

    modality_urls: ClassVar[dict[str, set[str]]] = {
        'esa_wc': {'esa_wc/ESA_WC.zip'},
        'viirs': {'viirs/VIIRS.zip'},
        'mask': {'mask/mask.zip'},
        's1': {'s1/s1_part1.zip', 's1/s1_part2.zip'},
        's2_temporal_subset': {
            's2_temporal_subset/s2_temporal_subset_part1.zip',
            's2_temporal_subset/s2_temporal_subset_part2.zip',
        },
        's2_autumn': {'s2_autumn/s2_autumn_part1.zip', 's2_autumn/s2_autumn_part2.zip'},
        's2_spring': {'s2_spring/s2_spring_part1.zip', 's2_spring/s2_spring_part2.zip'},
        's2_summer': {'s2_summer/s2_summer_part1.zip', 's2_summer/s2_summer_part2.zip'},
        's2_winter': {'s2_winter/s2_winter_part1.zip', 's2_winter/s2_winter_part2.zip'},
        'split_IDs': {'split_IDs/split_IDs.csv'},
    }

    md5s: ClassVar[dict[str, str]] = {
        'ESA_WC.zip': '72b2ee578fe10f0df85bdb7f19311c92',
        'VIIRS.zip': '4eff014bae127fe536f8a5f17d89ecb4',
        'mask.zip': '87c83a23a73998ad60d448d240b66225',
        's1_part1.zip': 'd8a911f5c76b50eb0760b8f0047e4674',
        's1_part2.zip': 'a30369d17c62d2af5aa52a4189590e3c',
        's2_temporal_subset_part1.zip': '78c2d05514458a036fe133f1e2f11d2a',
        's2_temporal_subset_part2.zip': '076cd3bd00eb5b7f5d80c9e0a0de0275',
        's2_autumn_part1.zip': '6ee7d1ac44b5107e3663636269aecf68',
        's2_autumn_part2.zip': '4fc5e1d5c772421dba553722433ac3b9',
        's2_spring_part1.zip': '2a89687d8fafa7fc7f5e641bfa97d472',
        's2_spring_part2.zip': '5845dcae0ab3cdc174b7c41edd4283a9',
        's2_summer_part1.zip': '73ca8291d3f4fb7533636220a816bb71',
        's2_summer_part2.zip': '5b5816bbd32987619bf72cde5cacd032',
        's2_winter_part1.zip': 'ca958f7cd98e37cb59d6f3877573ee6d',
        's2_winter_part2.zip': 'e7aacb0806d6d619b6abc408e6b09fdc',
        'split_IDs.csv': 'cb5c6c073702acee23544e1e6fe5856f',
    }

    mask_cmap: ClassVar[dict[int, tuple[int, int, int]]] = {
        1: (0, 153, 0),
        0: (255, 255, 255),
    }

    wc_cmap: ClassVar[dict[int, tuple[int, int, int]]] = {
        10: (0, 160, 0),
        20: (150, 100, 0),
        30: (255, 180, 0),
        40: (255, 255, 100),
        50: (195, 20, 0),
        60: (255, 245, 215),
        70: (255, 255, 255),
        80: (0, 70, 200),
        90: (0, 220, 130),
        95: (0, 150, 120),
        100: (255, 235, 175),
    }

    def __init__(
        self,
        root: Path = 'data',
        modality: list[str] = ['mask', 'esa_wc', 'viirs', 's2_summer'],
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new MapInWild dataset instance.

        Args:
            root: root directory where dataset can be found
            modality: the modality to download. Choose from: "mask", "esa_wc",
                "viirs", "s1", "s2_temporal_subset", "s2_[season]".
            split: one of "train", "validation", or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if ``split`` argument is invalid
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in ['train', 'validation', 'test']

        self.checksum = checksum
        self.root = root
        self.transforms = transforms
        self.modality = modality
        self.download = download

        modality.append('split_IDs')
        for mode in modality:
            for modality_link in self.modality_urls[mode]:
                modality_url = os.path.join(self.url, modality_link)
                self._verify(
                    url=modality_url, md5=self.md5s[os.path.split(modality_link)[-1]]
                )

            # Merge modalities downloaded in two parts
            if (
                download
                and mode not in os.listdir(self.root)
                and len(self.modality_urls[mode]) == 2
            ):
                self._merge_parts(mode)

        # Masks will be loaded separately in the :meth:`__getitem__`
        if 'mask' in self.modality:
            self.modality.remove('mask')

        # Split IDs has been downloaded and is not needed in the list
        if 'split_IDs' in self.modality:
            self.modality.remove('split_IDs')

        if os.path.exists(os.path.join(self.root, 'split_IDs.csv')):
            split_dataframe = pd.read_csv(os.path.join(self.root, 'split_IDs.csv'))
            self.ids = split_dataframe[split].dropna().values.tolist()
            self.ids = list(map(int, self.ids))

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        list_modalities = []
        id = self.ids[index]

        mask = self._load_raster(id, 'mask')
        mask[mask != 0] = 1

        for mode in self.modality:
            mode = mode.upper() if mode in ['esa_wc', 'viirs'] else mode
            data = self._load_raster(id, mode)
            list_modalities.append(data)

        image = torch.cat(list_modalities, dim=0)

        sample: dict[str, Tensor] = {'image': image, 'mask': mask}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.ids)

    def _load_raster(self, filename: int, source: Path) -> Tensor:
        """Load a single raster image or target.

        Args:
            filename: name of the file to load
            source: the directory of the modality

        Returns:
            the raster image or target
        """
        with rasterio.open(os.path.join(self.root, source, f'{filename}.tif')) as f:
            raw_array = f.read()
            array: np.typing.NDArray[np.int_] = np.stack(raw_array, axis=0)
            if array.dtype == np.uint16:
                array = array.astype(np.int32)
            tensor = torch.from_numpy(array).float()
            return tensor

    def _verify(self, url: str, md5: str | None = None) -> None:
        """Verify the integrity of the dataset.

        Args:
            url: url to the file
            md5: md5 of the file to be verified
        """
        modality_folder_name = url.split('/')[-1]
        mod_fold_no_ext = modality_folder_name.split('.')[0]
        modality_path = os.path.join(self.root, mod_fold_no_ext)
        split_path = os.path.join(self.root, modality_folder_name)
        if mod_fold_no_ext == 'split_IDs':
            modality_path = split_path

        # Check if the files already exist
        if os.path.exists(modality_path):
            return

        # Check if the zip files have already been downloaded, if so, extract
        filepath = os.path.join(self.root, url.split('/')[-1])
        if os.path.isfile(filepath) and filepath.endswith('.zip'):
            if self.checksum and not check_integrity(filepath, md5):
                raise RuntimeError('Dataset found, but corrupted.')
            self._extract(url)
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download(url, md5)
        if not url.endswith('.csv'):
            self._extract(url)

    def _download(self, url: str, md5: str | None) -> None:
        """Downloads a modality.

        Args:
            url: download url of a modality
            md5: md5 of a modality
        """
        download_url(
            url,
            self.root,
            filename=os.path.split(url)[1],
            md5=md5 if self.checksum else None,
        )

    def _extract(self, path: Path) -> None:
        """Extracts a modality.

        Args:
            path: path to the modality folder
        """
        filepath = os.path.join(self.root, os.path.split(path)[1])
        extract_archive(filepath)

    def _merge_parts(self, modality: str) -> None:
        """Merge the modalities that are downloaded and extracted in two parts.

        Args:
            root: root directory where dataset can be found
            modality: the filename of the modality
        """
        # Create a new folder named after the 'modality' variable
        modality_folder = os.path.join(self.root, modality)
        # Will not raise an error if the folder already exists
        os.makedirs(modality_folder, exist_ok=True)

        # List of source folders
        source_folders = [
            os.path.join(self.root, modality + '_part1'),
            os.path.join(self.root, modality + '_part2'),
        ]

        # Move files from each source folder to the new 'modality' folder
        for source_folder in source_folders:
            for file_name in os.listdir(source_folder):
                source = os.path.join(source_folder, file_name)
                destination = os.path.join(modality_folder, file_name)
                if os.path.isfile(source):
                    shutil.copy(source, destination)  # Move files to 'modality' folder

    def _convert_to_color(
        self, arr_2d: Tensor, cmap: dict[int, tuple[int, int, int]]
    ) -> 'np.typing.NDArray[np.uint8]':
        """Numeric labels to RGB-color encoding.

        Args:
            arr_2d: 2D array to be colorized
            cmap: colormap to use when mapping the labels

        Returns:
            3D colored image
        """
        arr_3d = np.zeros((arr_2d.shape[0], arr_2d.shape[1], 3), dtype=np.uint8)

        for c, i in cmap.items():
            m = arr_2d == c
            arr_3d[m] = i
        return arr_3d

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample image-mask pair returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        modality_channels = defaultdict(lambda: 10, {'viirs': 1, 'esa_wc': 1, 's1': 2})

        start_idx = 0
        split_images = {}

        for modality in self.modality:
            end_idx = start_idx + modality_channels[modality]  # Start + n of channels
            split_images[modality] = sample['image'][start_idx:end_idx, :, :]  # Slicing
            start_idx = end_idx  # Update the iterator

        # Prepare the mask
        mask = sample['mask'].squeeze()
        color_mask = self._convert_to_color(mask, cmap=self.mask_cmap)

        num_subplots = len(split_images) + 1  # +1 for color_mask
        showing_predictions = 'prediction' in sample
        if showing_predictions:
            num_subplots += 1

        fig, axs = plt.subplots(1, num_subplots, figsize=(num_subplots * 4, 5))

        # Plot each modality in its respective axis
        for i, (modality, image) in enumerate(split_images.items()):
            ax = axs[i]
            img = np.transpose(image, (1, 2, 0)).squeeze()
            # Apply transformations based on modality type
            if modality.startswith('s2'):
                img = img[:, :, [4, 3, 2]]
            if modality == 'esa_wc':
                img = self._convert_to_color(torch.as_tensor(img), cmap=self.wc_cmap)
            if modality == 's1':
                img = img[:, :, 0]

            if not 'esa_wc':
                img = percentile_normalization(img)

            ax.imshow(img)
            if show_titles:
                ax.set_title(modality)
            ax.axis('off')

        # Plot color_mask in its own axis
        axs[len(split_images)].imshow(color_mask)
        if show_titles:
            axs[len(split_images)].set_title('Annotation')
        axs[len(split_images)].axis('off')

        # If available, plot predictions in a new axis
        if showing_predictions:
            prediction = sample['prediction'].squeeze()
            color_predictions = self._convert_to_color(prediction, cmap=self.mask_cmap)
            axs[-1].imshow(color_predictions, vmin=0, vmax=1, interpolation='none')
            if show_titles:
                axs[-1].set_title('Prediction')
            axs[-1].axis('off')

        plt.tight_layout()
        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""USAVars datamodule."""

from typing import Any

from ..datasets import USAVars
from .geo import NonGeoDataModule


class USAVarsDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the USAVars dataset.

    Uses random train/val/test splits.

    .. versionadded:: 0.3
    """

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new USAVarsDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.USAVars`.
        """
        super().__init__(USAVars, batch_size, num_workers, **kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SatlasPretrain dataset."""

import os
from collections.abc import Callable, Iterable
from typing import ClassVar, TypedDict

import numpy as np
import pandas as pd
import torch
from einops import rearrange
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from PIL import Image
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, check_integrity, extract_archive, which


class _Task(TypedDict, total=False):
    BackgroundInvalid: bool
    categories: list[str]
    colors: list[list[int]]
    type: str


# https://github.com/allenai/satlas/blob/main/satlas/model/dataset.py
TASKS: dict[str, _Task] = {
    'polyline_bin_segment': {
        'type': 'bin_segment',
        'categories': [
            'airport_runway',
            'airport_taxiway',
            'raceway',
            'road',
            'railway',
            'river',
        ],
        'colors': [
            [255, 255, 255],  # (white) airport_runway
            [192, 192, 192],  # (light grey) airport_taxiway
            [160, 82, 45],  # (sienna) raceway
            [255, 255, 255],  # (white) road
            [144, 238, 144],  # (light green) railway
            [0, 0, 255],  # (blue) river
        ],
    },
    'bin_segment': {
        'type': 'bin_segment',
        'categories': [
            'aquafarm',
            'lock',
            'dam',
            'solar_farm',
            'power_plant',
            'gas_station',
            'park',
            'parking_garage',
            'parking_lot',
            'landfill',
            'quarry',
            'stadium',
            'airport',
            'airport_runway',
            'airport_taxiway',
            'airport_apron',
            'airport_hangar',
            'airstrip',
            'airport_terminal',
            'ski_resort',
            'theme_park',
            'storage_tank',
            'silo',
            'track',
            'raceway',
            'wastewater_plant',
            'road',
            'railway',
            'river',
            'water_park',
            'pier',
            'water_tower',
            'street_lamp',
            'traffic_signals',
            'power_tower',
            'power_substation',
            'building',
            'bridge',
            'road_motorway',
            'road_trunk',
            'road_primary',
            'road_secondary',
            'road_tertiary',
            'road_residential',
            'road_service',
            'road_track',
            'road_pedestrian',
        ],
        'colors': [
            [32, 178, 170],  # (light sea green) aquafarm
            [0, 255, 255],  # (cyan) lock
            [173, 216, 230],  # (light blue) dam
            [255, 0, 255],  # (magenta) solar farm
            [255, 165, 0],  # (orange) power plant
            [128, 128, 0],  # (olive) gas station
            [0, 255, 0],  # (green) park
            [47, 79, 79],  # (dark slate gray) parking garage
            [128, 0, 0],  # (maroon) parking lot
            [165, 42, 42],  # (brown) landfill
            [128, 128, 128],  # (grey) quarry
            [255, 215, 0],  # (gold) stadium
            [255, 105, 180],  # (pink) airport
            [255, 255, 255],  # (white) airport_runway
            [192, 192, 192],  # (light grey) airport_taxiway
            [128, 0, 128],  # (purple) airport_apron
            [0, 128, 0],  # (dark green) airport_hangar
            [248, 248, 255],  # (ghost white) airstrip
            [240, 230, 140],  # (khaki) airport_terminal
            [192, 192, 192],  # (silver) ski_resort
            [0, 96, 0],  # (dark green) theme_park
            [95, 158, 160],  # (cadet blue) storage_tank
            [205, 133, 63],  # (peru) silo
            [154, 205, 50],  # (yellow green) track
            [160, 82, 45],  # (sienna) raceway
            [218, 112, 214],  # (orchid) wastewater_plant
            [255, 255, 255],  # (white) road
            [144, 238, 144],  # (light green) railway
            [0, 0, 255],  # (blue) river
            [255, 240, 245],  # (lavender blush) water_park
            [65, 105, 225],  # (royal blue) pier
            [238, 130, 238],  # (violet) water_tower
            [75, 0, 130],  # (indigo) street_lamp
            [233, 150, 122],  # (dark salmon) traffic_signals
            [255, 255, 0],  # (yellow) power_tower
            [255, 255, 0],  # (yellow) power_substation
            [255, 0, 0],  # (red) building
            [64, 64, 64],  # (dark grey) bridge
            [255, 255, 255],  # (white) road_motorway
            [255, 255, 255],  # (white) road_trunk
            [255, 255, 255],  # (white) road_primary
            [255, 255, 255],  # (white) road_secondary
            [255, 255, 255],  # (white) road_tertiary
            [255, 255, 255],  # (white) road_residential
            [255, 255, 255],  # (white) road_service
            [255, 255, 255],  # (white) road_track
            [255, 255, 255],  # (white) road_pedestrian
        ],
    },
    'land_cover': {
        'type': 'segment',
        'BackgroundInvalid': True,
        'categories': [
            'background',
            'water',
            'developed',
            'tree',
            'shrub',
            'grass',
            'crop',
            'bare',
            'snow',
            'wetland',
            'mangroves',
            'moss',
        ],
        'colors': [
            [0, 0, 0],  # unknown
            [0, 0, 255],  # (blue) water
            [255, 0, 0],  # (red) developed
            [0, 192, 0],  # (dark green) tree
            [200, 170, 120],  # (brown) shrub
            [0, 255, 0],  # (green) grass
            [255, 255, 0],  # (yellow) crop
            [128, 128, 128],  # (grey) bare
            [255, 255, 255],  # (white) snow
            [0, 255, 255],  # (cyan) wetland
            [255, 0, 255],  # (pink) mangroves
            [128, 0, 128],  # (purple) moss
        ],
    },
    'tree_cover': {'type': 'regress', 'BackgroundInvalid': True},
    'crop_type': {
        'type': 'segment',
        'BackgroundInvalid': True,
        'categories': [
            'invalid',
            'rice',
            'grape',
            'corn',
            'sugarcane',
            'tea',
            'hop',
            'wheat',
            'soy',
            'barley',
            'oats',
            'rye',
            'cassava',
            'potato',
            'sunflower',
            'asparagus',
            'coffee',
        ],
        'colors': [
            [0, 0, 0],  # unknown
            [0, 0, 255],  # (blue) rice
            [255, 0, 0],  # (red) grape
            [255, 255, 0],  # (yellow) corn
            [0, 255, 0],  # (green) sugarcane
            [128, 0, 128],  # (purple) tea
            [255, 0, 255],  # (pink) hop
            [0, 128, 0],  # (dark green) wheat
            [255, 255, 255],  # (white) soy
            [128, 128, 128],  # (grey) barley
            [165, 42, 42],  # (brown) oats
            [0, 255, 255],  # (cyan) rye
            [128, 0, 0],  # (maroon) cassava
            [173, 216, 230],  # (light blue) potato
            [128, 128, 0],  # (olive) sunflower
            [0, 128, 0],  # (dark green) asparagus
            [92, 64, 51],  # (dark brown) coffee
        ],
    },
    'point': {
        'type': 'detect',
        'categories': [
            'background',
            'wind_turbine',
            'lighthouse',
            'mineshaft',
            'aerialway_pylon',
            'helipad',
            'fountain',
            'toll_booth',
            'chimney',
            'communications_tower',
            'flagpole',
            'petroleum_well',
            'water_tower',
            'offshore_wind_turbine',
            'offshore_platform',
            'power_tower',
        ],
        'colors': [
            [0, 0, 0],
            [0, 255, 255],  # (cyan) wind_turbine
            [0, 255, 0],  # (green) lighthouse
            [255, 255, 0],  # (yellow) mineshaft
            [0, 0, 255],  # (blue) pylon
            [173, 216, 230],  # (light blue) helipad
            [128, 0, 128],  # (purple) fountain
            [255, 255, 255],  # (white) toll_booth
            [0, 128, 0],  # (dark green) chimney
            [128, 128, 128],  # (grey) communications_tower
            [165, 42, 42],  # (brown) flagpole
            [128, 0, 0],  # (maroon) petroleum_well
            [255, 165, 0],  # (orange) water_tower
            [255, 255, 0],  # (yellow) offshore_wind_turbine
            [255, 0, 0],  # (red) offshore_platform
            [255, 0, 255],  # (magenta) power_tower
        ],
    },
    'rooftop_solar_panel': {
        'type': 'detect',
        'categories': ['background', 'rooftop_solar_panel'],
        'colors': [
            [0, 0, 0],
            [255, 255, 0],  # (yellow) rooftop_solar_panel
        ],
    },
    'building': {
        'type': 'instance',
        'categories': ['background', 'ms_building'],
        'colors': [
            [0, 0, 0],
            [255, 255, 0],  # (yellow) building
        ],
    },
    'polygon': {
        'type': 'instance',
        'categories': [
            'background',
            'aquafarm',
            'lock',
            'dam',
            'solar_farm',
            'power_plant',
            'gas_station',
            'park',
            'parking_garage',
            'parking_lot',
            'landfill',
            'quarry',
            'stadium',
            'airport',
            'airport_apron',
            'airport_hangar',
            'airport_terminal',
            'ski_resort',
            'theme_park',
            'storage_tank',
            'silo',
            'track',
            'wastewater_plant',
            'power_substation',
            'pier',
            'crop',
            'water_park',
        ],
        'colors': [
            [0, 0, 0],
            [255, 255, 0],  # (yellow) aquafarm
            [0, 255, 255],  # (cyan) lock
            [0, 255, 0],  # (green) dam
            [0, 0, 255],  # (blue) solar_farm
            [255, 0, 0],  # (red) power_plant
            [128, 0, 128],  # (purple) gas_station
            [255, 255, 255],  # (white) park
            [0, 128, 0],  # (dark green) parking_garage
            [128, 128, 128],  # (grey) parking_lot
            [165, 42, 42],  # (brown) landfill
            [128, 0, 0],  # (maroon) quarry
            [255, 165, 0],  # (orange) stadium
            [255, 105, 180],  # (pink) airport
            [192, 192, 192],  # (silver) airport_apron
            [173, 216, 230],  # (light blue) airport_hangar
            [32, 178, 170],  # (light sea green) airport_terminal
            [255, 0, 255],  # (magenta) ski_resort
            [128, 128, 0],  # (olive) theme_park
            [47, 79, 79],  # (dark slate gray) storage_tank
            [255, 215, 0],  # (gold) silo
            [192, 192, 192],  # (light grey) track
            [240, 230, 140],  # (khaki) wastewater_plant
            [154, 205, 50],  # (yellow green) power_substation
            [255, 165, 0],  # (orange) pier
            [0, 192, 0],  # (middle green) crop
            [0, 192, 0],  # (middle green) water_park
        ],
    },
    'wildfire': {
        'type': 'bin_segment',
        'categories': ['fire_retardant', 'burned'],
        'colors': [
            [255, 0, 0],  # (red) fire retardant
            [128, 128, 128],  # (grey) burned area
        ],
    },
    'smoke': {'type': 'classification', 'categories': ['no', 'partial', 'yes']},
    'snow': {'type': 'classification', 'categories': ['no', 'partial', 'yes']},
    'dem': {'type': 'regress', 'BackgroundInvalid': True},
    'airplane': {
        'type': 'detect',
        'categories': ['background', 'airplane'],
        'colors': [
            [0, 0, 0],  # (black) background
            [255, 0, 0],  # (red) airplane
        ],
    },
    'vessel': {
        'type': 'detect',
        'categories': ['background', 'vessel'],
        'colors': [
            [0, 0, 0],  # (black) background
            [255, 0, 0],  # (red) vessel
        ],
    },
    'water_event': {
        'type': 'segment',
        'BackgroundInvalid': True,
        'categories': ['invalid', 'background', 'water_event'],
        'colors': [
            [0, 0, 0],  # (black) invalid
            [0, 255, 0],  # (green) background
            [0, 0, 255],  # (blue) water_event
        ],
    },
    'park_sport': {
        'type': 'classification',
        'categories': [
            'american_football',
            'badminton',
            'baseball',
            'basketball',
            'cricket',
            'rugby',
            'soccer',
            'tennis',
            'volleyball',
        ],
    },
    'park_type': {
        'type': 'classification',
        'categories': ['park', 'pitch', 'golf_course', 'cemetery'],
    },
    'power_plant_type': {
        'type': 'classification',
        'categories': ['oil', 'nuclear', 'coal', 'gas'],
    },
    'quarry_resource': {
        'type': 'classification',
        'categories': ['sand', 'gravel', 'clay', 'coal', 'peat'],
    },
    'track_sport': {
        'type': 'classification',
        'categories': ['running', 'cycling', 'horse'],
    },
    'road_type': {
        'type': 'classification',
        'categories': [
            'motorway',
            'trunk',
            'primary',
            'secondary',
            'tertiary',
            'residential',
            'service',
            'track',
            'pedestrian',
        ],
    },
    'cloud': {
        'type': 'bin_segment',
        'categories': ['background', 'cloud', 'shadow'],
        'colors': [
            [0, 255, 0],  # (green) not clouds or shadows
            [255, 255, 255],  # (white) clouds
            [128, 128, 128],  # (grey) shadows
        ],
        'BackgroundInvalid': True,
    },
    'flood': {
        'type': 'bin_segment',
        'categories': ['background', 'water'],
        'colors': [
            [0, 255, 0],  # (green) background
            [0, 0, 255],  # (blue) water
        ],
        'BackgroundInvalid': True,
    },
}


class SatlasPretrain(NonGeoDataset):
    """SatlasPretrain dataset.

    `SatlasPretrain <https://satlas-pretrain.allen.ai/>`__ is a large-scale pre-training
    dataset for tasks that involve understanding satellite images. Regularly-updated
    satellite data is publicly available for much of the Earth through sources such as
    Sentinel-2 and NAIP, and can inform numerous applications from tackling illegal
    deforestation to monitoring marine infrastructure. However, developing automatic
    computer vision systems to parse these images requires a huge amount of manual
    labeling of training data. By combining over 30 TB of satellite images with 137
    label categories, SatlasPretrain serves as an effective pre-training dataset that
    greatly reduces the effort needed to develop robust models for downstream satellite
    image applications.

    Reference implementation:

    * https://github.com/allenai/satlas/blob/main/satlas/model/dataset.py

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.48550/arXiv.2211.15660

    .. versionadded:: 0.7

    .. note::
       This dataset requires the following additional library to be installed:

       * `AWS CLI <https://aws.amazon.com/cli/>`_: to download the dataset from AWS.
    """

    # https://github.com/allenai/satlas/blob/main/satlaspretrain_urls.txt
    url = 's3://ai2-public-datasets/satlas/'
    tarballs: ClassVar[dict[str, tuple[str, ...]]] = {
        'landsat': ('satlas-dataset-v1-landsat.tar',),
        'naip': (
            'satlas-dataset-v1-naip-2011.tar',
            'satlas-dataset-v1-naip-2012.tar',
            'satlas-dataset-v1-naip-2013.tar',
            'satlas-dataset-v1-naip-2014.tar',
            'satlas-dataset-v1-naip-2015.tar',
            'satlas-dataset-v1-naip-2016.tar',
            'satlas-dataset-v1-naip-2017.tar',
            'satlas-dataset-v1-naip-2018.tar',
            'satlas-dataset-v1-naip-2019.tar',
            'satlas-dataset-v1-naip-2020.tar',
        ),
        'sentinel1': ('satlas-dataset-v1-sentinel1-new.tar',),
        'sentinel2': (
            'satlas-dataset-v1-sentinel2-a.tar',
            'satlas-dataset-v1-sentinel2-b.tar',
        ),
        'static': ('satlas-dataset-v1-labels-static.tar',),
        'dynamic': ('satlas-dataset-v1-labels-dynamic.tar',),
        'metadata': ('satlas-dataset-v1-metadata.tar',),
    }
    md5s: ClassVar[dict[str, tuple[str, ...]]] = {
        'landsat': ('89ea5e8974826c071908392827780a06',),
        'naip': (
            '523736842994861054f04b97c4d90bfb',
            '636b9a3b08be0e40d098cb7b5e655b57',
            '69e2b1052b1d2d465322a24cf7207a16',
            '38999aea424d403ad60e1398443636aa',
            '97f4855072a8a406a4bfbe94c5f7311c',
            '9ba3c626b23e6d26749a323eaedc7c0a',
            'e4aba3d198dedfe1524a9338e85794aa',
            '74191a36d841b0b9b5d5cbae9a92ad71',
            '55b110cc6f734bf88793306d49f1c415',
            '97fc8414334987c59593d574f112a77e',
        ),
        'sentinel1': ('3d88a0a10df6ab0aa50db2ba4c475048',),
        'sentinel2': (
            '7e1c6a1e322807fb11df8c0c062545ca',
            '6636b8ecf2fff1d6723ecfef55a4876d',
        ),
        'static': ('4e38c2573bc78cf1f0d7267e432cb42c',),
        'dynamic': ('4503ae687948e7d2cb7ade0083f77a8a',),
        'metadata': ('6b9ac5a4f9a1ee88a271d28f12854607',),
    }

    # NOTE: 'tci' is RGB (b04-b02), not BGR (b02-b04)
    bands: ClassVar[dict[str, tuple[str, ...]]] = {
        'landsat': tuple(f'b{i}' for i in range(1, 12)),
        'naip': ('tci', 'ir'),
        'sentinel1': ('vh', 'vv'),
        'sentinel2': ('tci', 'b05', 'b06', 'b07', 'b08', 'b11', 'b12'),
    }

    chip_size = 512

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train_lowres',
        good_images: str = 'good_images_lowres_all',
        image_times: str = 'image_times',
        images: Iterable[str] = ('sentinel1', 'sentinel2', 'landsat'),
        labels: Iterable[str] = ('land_cover',),
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new SatlasPretrain instance.

        Args:
            root: Root directory where dataset can be found.
            split: Metadata split to load.
            good_images: Metadata mapping between col/row and directory.
            image_times: Metadata mapping between directory and ISO time.
            images: List of image products.
            labels: List of label products.
            transforms: A function/transform that takes input sample and its target as
                entry and returns a transformed version.
            download: If True, download dataset and store it in the root directory.
            checksum: If True, check the MD5 of the downloaded files (may be slow).

        Raises:
            AssertionError: If *images* is invalid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert set(images) <= set(self.bands.keys())

        self.root = root
        self.images = images
        self.labels = labels
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        # Read metadata files
        self.split = pd.read_json(
            os.path.join(root, 'metadata', f'{split}.json'), typ='frame'
        )
        self.good_images = pd.read_json(
            os.path.join(root, 'metadata', f'{good_images}.json'), typ='frame'
        )
        self.image_times = pd.read_json(
            os.path.join(root, 'metadata', f'{image_times}.json'), typ='series'
        )

        self.split.columns = ['col', 'row']
        self.good_images.columns = ['col', 'row', 'directory']
        self.good_images = self.good_images.groupby(['col', 'row'])

    def __len__(self) -> int:
        """Return the number of locations in the dataset.

        Returns:
            Length of the dataset
        """
        return len(self.split)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and label at that index.
        """
        col, row = self.split.iloc[index]
        directories = self.good_images.get_group((col, row))['directory']

        sample: dict[str, Tensor] = {}

        for image in self.images:
            self._load_image(sample, image, col, row, directories)

        for label in self.labels:
            self._load_label(sample, label, col, row)

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _load_image(
        self,
        sample: dict[str, Tensor],
        image: str,
        col: int,
        row: int,
        directories: pd.Series,
    ) -> None:
        """Load a single image.

        Args:
            sample: Dataset sample to populate.
            image: Image product.
            col: Web Mercator column.
            row: Web Mercator row.
            directories: Directories that may contain the image.
        """
        # Find directories that match image product
        good_directories: list[str] = []
        for directory in directories:
            path = os.path.join(self.root, image, directory)
            if os.path.isdir(path):
                good_directories.append(directory)

        # Choose a random timestamp
        idx = torch.randint(len(good_directories), (1,))
        directory = good_directories[idx]
        time = self.image_times[directory].timestamp()
        sample[f'time_{image}'] = torch.tensor(time)

        # Load all bands
        resample = Image.Resampling.BILINEAR
        channels = []
        for band in self.bands[image]:
            path = os.path.join(self.root, image, directory, band, f'{col}_{row}.png')
            with Image.open(path) as img:
                img = img.resize((self.chip_size, self.chip_size), resample=resample)
                array = np.atleast_3d(np.array(img, dtype=np.float32))
                channels.append(torch.tensor(array))
        raster = rearrange(torch.cat(channels, dim=-1), 'h w c -> c h w')
        sample[f'image_{image}'] = raster

    def _load_label(
        self, sample: dict[str, Tensor], label: str, col: int, row: int
    ) -> None:
        """Load a single label.

        Args:
            sample: Dataset sample to populate.
            label: Label product.
            col: Web Mercator column.
            row: Web Mercator row.
        """
        path = os.path.join(self.root, 'static', f'{col}_{row}', f'{label}.png')
        if os.path.isfile(path):
            with Image.open(path) as img:
                raster = torch.tensor(np.array(img, dtype=np.int64))
        else:
            raster = torch.zeros(self.chip_size, self.chip_size, dtype=torch.long)
        sample[f'mask_{label}'] = raster

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        products = [*self.images, 'metadata']
        if self.labels:
            products.append('static')

        for product in products:
            # Check if the extracted directory already exists
            if os.path.isdir(os.path.join(self.root, product)):
                continue

            tarballs = self.tarballs[product]
            md5s = self.md5s[product]
            for tarball, md5 in zip(tarballs, md5s):
                path = os.path.join(self.root, tarball)

                # Check if the tarball has already been downloaded
                if os.path.isfile(path):
                    extract_archive(path)
                    continue

                # Check if the user requested to download the dataset
                if not self.download:
                    raise DatasetNotFoundError(self)

                # Download and extract the tarball
                aws = which('aws')
                aws('s3', 'cp', self.url + tarball, self.root)
                check_integrity(path, md5 if self.checksum else None)
                extract_archive(path)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: A sample returned by :meth:`__getitem__`.
            show_titles: Flag indicating whether to show titles above each panel.
            suptitle: Optional string to use as a suptitle.

        Returns:
            A matplotlib Figure with the rendered sample.
        """
        images = []
        titles = []
        for key, value in sample.items():
            match key.split('_', 1):
                case ['image', 'landsat']:
                    images.append(rearrange(value[[3, 2, 1]], 'c h w -> h w c') / 255)
                    titles.append('Landsat 8/9')
                case ['image', 'naip']:
                    images.append(rearrange(value[:3], 'c h w -> h w c') / 255)
                    titles.append('NAIP')
                case ['image', 'sentinel1']:
                    images.extend([value[0] / 255, value[1] / 255])
                    titles.extend(['Sentinel-1 VH', 'Sentinel-1 VV'])
                case ['image', 'sentinel2']:
                    images.append(rearrange(value[:3], 'c h w -> h w c') / 255)
                    titles.append('Sentinel-2')
                case ['mask' | 'prediction', label]:
                    cmap = torch.tensor(TASKS[label]['colors'])
                    images.append(cmap[value])
                    titles.append(label.replace('_', ' ').capitalize())

        fig, ax = plt.subplots(ncols=len(images), squeeze=False)
        for i, (image, title) in enumerate(zip(images, titles)):
            ax[0, i].imshow(image)
            ax[0, i].axis('off')

            if show_titles:
                ax[0, i].set_title(title)

        if suptitle is not None:
            fig.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Copernicus-Bench LCZ-S2 dataset."""

import os
from collections.abc import Callable, Sequence
from typing import ClassVar, Literal

import torch
from einops import rearrange
from torch import Tensor

from ..errors import DatasetNotFoundError
from ..utils import Path, download_url, lazy_import
from .base import CopernicusBenchBase


class CopernicusBenchLCZS2(CopernicusBenchBase):
    """Copernicus-Bench LCZ-S2 dataset.

    LCZ-S2 is a multi-class scene classification dataset derived from So2Sat-LCZ42,
    a large-scale local climate zone classification dataset.

    If you use this dataset in your research, please cite the following papers:

    * https://arxiv.org/abs/2503.11849
    * https://doi.org/10.1109/MGRS.2020.2964708

    .. note::

       This dataset requires the following additional library to be installed:

       * `<https://pypi.org/project/h5py/>`_ to load the dataset.

    .. versionadded:: 0.7
    """

    url = 'https://hf.co/datasets/wangyi111/Copernicus-Bench/resolve/9d252acd3aa0e3da3128e05c6f028647f0e48e5f/l3_lcz_s2/lcz_{}.h5'
    md5s: ClassVar[dict[str, str]] = {
        'train': 'e0b10cdb7f12e053cda8dd3ff12dbd9e',
        'val': 'be3b503dba5a1405ec6d5a770c2bee33',
        'test': '4e95788c72a421d636f6f8dc7623d116',
    }
    filename = 'lcz_{}.h5'
    all_bands = ('B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B11', 'B12')
    rgb_bands = ('B04', 'B03', 'B02')
    classes = (
        'Compact high rise',
        'Compact mid rise',
        'Compact low rise',
        'Open high rise',
        'Open mid rise',
        'Open low rise',
        'Lightweight low rise',
        'Large low rise',
        'Sparsely built',
        'Heavy industry',
        'Dense trees',
        'Scattered trees',
        'Bush, scrub',
        'Low plants',
        'Bare rock or paved',
        'Bare soil or sand',
        'Water',
    )

    def __init__(
        self,
        root: Path = 'data',
        split: Literal['train', 'val', 'test'] = 'train',
        bands: Sequence[str] | None = None,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new CopernicusBenchBase instance.

        Args:
            root: Root directory where dataset can be found.
            split: One of 'train', 'val', or 'test'.
            bands: Sequence of band names to load (defaults to all bands).
            transforms: A function/transform that takes input sample and its target as
                entry and returns a transformed version.
            download: If True, download dataset and store it in the root directory.
            checksum: If True, check the MD5 of the downloaded files (may be slow).

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        h5py = lazy_import('h5py')

        self.root = root
        self.split = split
        self.bands = bands or self.all_bands
        self.band_indices = [self.all_bands.index(i) for i in self.bands]
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        self.filepath = os.path.join(root, self.filename.format(split))
        with h5py.File(self.filepath, 'r') as f:
            self.length: int = f['label'].shape[0]

    def __len__(self) -> int:
        """Return the length of the dataset.

        Returns:
            Length of the dataset.
        """
        return self.length

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: Index to return.

        Returns:
            Data and labels at that index.
        """
        h5py = lazy_import('h5py')

        with h5py.File(self.filepath, 'r') as f:
            sen2 = f['sen2'][index][:, :, self.band_indices]
            sen2 = rearrange(sen2, 'h w c -> c h w')
            label = f['label'][index].argmax()

        sample = {'image': torch.from_numpy(sen2), 'label': torch.tensor(label)}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        if os.path.exists(os.path.join(self.root, self.filename.format(self.split))):
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download and extract the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        md5 = self.md5s[self.split] if self.checksum else None
        download_url(self.url.format(self.split), self.root, md5=md5)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Tropical Cyclone Wind Estimation Competition datamodule."""

from typing import Any

from torch.utils.data import Subset

from ..datasets import TropicalCyclone
from .geo import NonGeoDataModule
from .utils import group_shuffle_split


class TropicalCycloneDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the NASA Cyclone dataset.

    Implements 80/20 train/val splits based on hurricane storm ids.
    See :func:`setup` for more details.

    .. versionchanged:: 0.4
        Class name changed from CycloneDataModule to TropicalCycloneDataModule to be
        consistent with TropicalCyclone dataset.
    """

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new TropicalCycloneDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.TropicalCyclone`.
        """
        super().__init__(TropicalCyclone, batch_size, num_workers, **kwargs)

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        if stage in ['fit', 'validate']:
            dataset = TropicalCyclone(split='train', **self.kwargs)
            train_indices, val_indices = group_shuffle_split(
                dataset.features['Storm ID'], test_size=0.2, random_state=0
            )
            self.train_dataset = Subset(dataset, train_indices)
            self.val_dataset = Subset(dataset, val_indices)
        if stage in ['test']:
            self.test_dataset = TropicalCyclone(split='test', **self.kwargs)
